"""
This module provides the paraview interface from homcloud.interface module.

The basic model of this module is "pipeline" model. A data object is
filtered by the chain of pipeline nodes, and the result is shown.
The objects of :class:`PipelineNode` correspond the node of the pipeline,
and you can adjust visualization by these objects.

You can use many methods of :class:`PipelineNode` to adjust the visualization,
and you can construct a new pipeline node by the following methods:

* :meth:`PipelineNode.threshold`
* :meth:`PipelineNode.clip_sphere`

Todo:

* Add clip_box method

"""
import numbers
import sys
from tempfile import NamedTemporaryFile
import os

import numpy as np
from forwardable import forwardable

import homcloud.utils as utils
import homcloud.pict.pict3d_vtk as pict3d_vtk
from homcloud.visualize_3d import ParaViewPolyLineDrawer, ParaView3DSurfaceDrawer, ParaViewLinesDrawer


@forwardable()
class TempFile(object):
    def __init__(self, suffix):
        self.named_tempfile = NamedTemporaryFile("w+", suffix=suffix, delete=False)
        self.cleanup_done = False

    def_delegators("named_tempfile", "name, close, write")

    def cleanup(self):
        self.close()
        if not self.cleanup_done:
            os.unlink(self.name)
            self.cleanup_done = True

    def __del__(self):
        self.cleanup()

    # NOTE: __enter__ and __exit__ are not implemented since
    # in this module with statement is used to the class


class PipelineNode(object):
    """This class represents elements in a pipeline in ParaView.

    This class is the base class for paraview interface classes.
    You should not create an instance of this class directly.
    """

    def __init__(self, parent=None):
        self.index = new_index()
        self.parent = parent
        self._representation = None
        self.opacity = None
        self.color = None
        self.color_field = None
        self.colorbar_range = None
        self.pointsize = None
        self.linewidth = None

    def variable(self):
        return "s" + str(self.index)

    def to_paraview_node(self):
        """Returns self."""
        return self

    def representation(self):
        if self._representation:
            return self._representation
        if self.parent is None:
            return False
        return self.parent.representation()

    def write_code(self, out, already_written, isleaf):
        if self.variable() in already_written:
            return
        already_written.add(self.variable())
        if self.parent:
            self.parent.write_code(out, already_written, False)
        self.write_constructor(out)
        if isleaf:
            out.write("Show({})\n".format(self.variable()))
        self.write_specialized_code(out, already_written)
        if isleaf:
            self.write_displayproperty_code(out, already_written)

    def write_displayproperty_code(self, out, already_written):
        out.write("dp = GetDisplayProperties({})\n".format(self.variable()))
        if self.representation():
            out.write("dp.Representation = '{}'\n".format(self.representation()))
        if self.opacity is not None:
            out.write("dp.Opacity = {}\n".format(self.opacity))
        if self.pointsize is not None:
            out.write("dp.PointSize = {}\n".format(self.pointsize))
        if self.color_field is not None:
            self.write_color_filed(out)
        if self.color is not None:
            out.write("dp.AmbientColor = ({}, {}, {})\n".format(*self.color))
            out.write("dp.DiffuseColor = ({}, {}, {})\n".format(*self.color))
            out.write("dp.ColorArrayName = [dp.ColorArrayName[0], r'']\n")
        if self.linewidth is not None:
            out.write("dp.LineWidth = {}\n".format(self.linewidth))

    def write_color_filed(self, out):
        if self.colorbar_range is None:
            out.write(
                "dp.LookupTable = MakeBlueToRedLT(*{}."
                "{}[r'{}'].GetRange())\n".format(self.variable(), self.data_attr_name(), self.color_field)
            )
        else:
            out.write("dp.LookupTable = MakeBlueToRedLT({}, {})\n".format(*self.colorbar_range))
        out.write("dp.ColorArrayName = [r'{}', r'{}']\n".format(self.color_array_type(), self.color_field))

    def write_specialized_code(self, out, already_written):
        pass

    def threshold(self, field, range):
        """
        Create a new pipeline node for thresholding whose parent
        is `self`.

        Only the elements whose `field` are in the `range` when
        the returned object is passed to :meth:`show`.

        Args:
            field (string): The name of the field which have the
                thresholded value.
            range (tuple[float, float] or float): The upper and lower bounds
                of the values.
                If an float number is given, the threshold is (range, range).

        Returns:
            Threshold: A new PipelineNode object for thresholding.

        """
        if isinstance(range, numbers.Number):
            range = (range, range)
        return Threshold(self, field, range)

    def clip_sphere(self, center, radius, inside_out=True):
        """
        Create a new pipeline node to clip the object with a
        sphere shape.

        Args:
            center (tuple[float, float, float]): The center of the
                clipping sphere
            radius (float): The radius of the clipping sphere
            inside_out (bool): If True, only the elements
                *in* the shpere are shown.
                If False, only the elements *outside* of the sphere is
                shown.

        Returns:
            SphereClip: A new PipelineNode object for sphere clipping.

        """
        return SphereClip(self, center, radius, inside_out)

    def color_by(self, field, range=None):
        """
        Set the coloring by field name.

        Args:
            field (string or int): The name of the field.
            range (tuple[float, float] or None): The upper and lower bounds
                of the colorbar. If None, the minimal and maximal values
                of the field are used.
        Returns:
            self
        """
        if isinstance(field, int):
            field = "Field {}".format(field)
        self.color_field = field
        self.colorbar_range = range
        return self

    def data_attr_name(self):
        return self.parent.data_attr_name()

    def color_array_type(self):
        return self.parent.color_array_type()

    def set_opacity(self, opacity):
        """
        Set the opacity.

        * 0.0 - completely transparent
        * 1.0 - completely opaque

        Args:
            opacity (float): The opacity.

        Returns:
            self
        """
        self.opacity = opacity
        return self

    def set_color(self, color):
        """
        Set the color.

        Args:
            color (tuple[float, float, float]): The RGB values (0.0 to 1.0)

        Returns:
            self
        """
        self.color = color
        return self

    def set_pointsize(self, size):
        """
        Set the pointsize.

        Args:
            pointsize (float): The size of the points

        Returns:
            self
        """
        self.pointsize = size
        return self

    def set_representation(self, rep):
        if rep not in ["Wireframe", "Points", "Surface", "Surface With Edges"]:
            raise ValueError("Representation {} is not acceptable".format(rep))
        self._representation = rep
        return self

    def set_linewidth(self, width):
        """
        Set the linewidth.

        Args:
            width (float): The width of the lines

        Returns:
            self
        """
        self.linewidth = width
        return self

    def debug_print(self):
        write_python_code(sys.stdout, [self])
        return self

    def cleanup(self):
        if self.parent:
            self.parent.cleanup()


class VTK(PipelineNode):
    """
    This class represents a VTK data source in paraview.
    """

    def __init__(self, path, gui_name=None, file=None):
        super().__init__()
        self.path = path
        self.gui_name = gui_name if gui_name is not None else path
        self.file = file

    def cleanup(self):
        if self.file:
            self.file.cleanup()

    def write_constructor(self, out):
        out.write(
            "{} = LegacyVTKReader(FileNames=r'{}',"
            " guiName=r'{}')\n".format(self.variable(), self.path, self.gui_name)
        )

    def data_attr_name(self):
        return "CellData"

    def color_array_type(self):
        return "CELLS"


class XMLVTI(PipelineNode):
    """
    This class represents a VTI data source in XML format in paraview.
    """

    def __init__(self, path, gui_name, file=None):
        super().__init__()
        self.path = path
        self.gui_name = gui_name if gui_name is not None else path
        self.file = file
        self._representation = "Surface"
        self.color_field = "value"

    def cleanup(self):
        if self.file:
            self.file.cleanup()

    def write_constructor(self, out):
        out.write(
            "{} = XMLImageDataReader(FileName=r'{}', "
            "guiName=r'{}')\n".format(self.variable(), self.path, self.gui_name)
        )

    def data_attr_name(self):
        return "CellData"

    def color_array_type(self):
        return "CELLS"


def VoxelData(array, gui_name=None, offsets=[0, 0, 0]):
    """
    Returns :class:`PipelineNode` object representing a voxel data.

    Args:
        array (numpy.ndarray): An array.
        gui_name (string or None): The name shown in Pipeline Browser
            in paraview's GUI.

    Returns:
        VTK: A pipeline node object
    """
    f = TempFile(".vti")
    pict3d_vtk.write_vti_xmlfile(f, array, offsets)
    f.close()
    return XMLVTI(f.name, gui_name, f)


def PolyLine(array, gui_name=None):
    """
    Returns :class:`PipelineNode` object representing a polyline.

    Args:
        array (numpy.ndarray): An array of points.
        gui_name (string or None): The name shown in Pipeline Browser
            in paraview's GUI.

    Returns:
        VTK: A pipeline node object
    """
    f = TempFile(".vtk")
    drawer = ParaViewPolyLineDrawer(2, array, {})
    drawer.write(f)
    f.close()
    return VTK(f.name, gui_name, f).set_representation("Wireframe")


def Lines(lines, gui_name=None):
    f = TempFile(".vtk")
    drawer = ParaViewLinesDrawer(1, {})
    for line in lines:
        drawer.draw_line(tuple(line[0]), tuple(line[1]), 0)
    drawer.write(f)
    f.close()
    return VTK(f.name, gui_name, f).set_representation("Wireframe")


def Surface(surface, gui_name=None):
    f = TempFile(".vtk")
    drawer = ParaView3DSurfaceDrawer(2, surface, {})
    drawer.write(f)
    f.close()
    return VTK(f.name, gui_name, f).set_representation("Surface")


class PointCloud(PipelineNode):
    """
    This class represents a pointcloud data source in paraview.

    Args:
        path (string): The filepath of the pointcloud.
        dim (int): The dimension of the space in which the pointcloud lives.
        delimiter (string): The delimiter of elements in a pointcloud file.
           If you want to show a CSV file, please specify ",".
        gui_name (string or None): The name shown in Pipeline Browser
            in paraview's GUI.

    Notes:
       This is constructed by CSVReader and TableToPoints.
    """

    def __init__(self, path, dim=3, delimiters=" ", gui_name=None, file=None):
        super().__init__()
        self.path = path
        self.dim = dim
        self.delimiters = delimiters
        self.gui_name = gui_name or self.path
        self.file = file
        assert dim in [2, 3]

    def cleanup(self):
        if self.file:
            self.file.cleanup()

    @staticmethod
    def from_array(array, dim=3, gui_name=None):
        """
        Construct a pipeline node for pointcloud from an ndarray object.

        Args:
            array (nupmy.ndarray): The pointcloud data.
            dim (int): The dimension of the space in which the
                pointcloud lives.
            gui_name (string): The name shown in Pipeline Browser in paraview's
                GUI.

        Returns:
            PointCloud: A pipeline node object.
        """
        f = TempFile(".txt")
        np.savetxt(f, array)
        f.close()
        return PointCloud(f.name, dim, " ", gui_name, f)

    def representation(self):
        return "Points"

    def write_constructor(self, out):
        def quote(s):
            return s.translate(
                {
                    "\t": "\\t",
                    "\n": "\\n",
                }
            )

        out.write(
            "s = CSVReader(FileName=r'{}', FieldDelimiterCharacters='{}',"
            " guiName=r'{}', HaveHeaders=0)\n".format(self.path, quote(self.delimiters), self.gui_name)
        )

        out.write(self.table_to_points_template().format(self.variable()))

    def table_to_points_template(self):
        if self.dim == 3:
            return "{} = TableToPoints(s, XColumn='Field 0', YColumn='Field 1', ZColumn='Field 2')\n"
        elif self.dim == 2:
            return "{} = TableToPoints(s, XColumn='Field 0', YColumn='Field 1', ZColumn='Field 0', a2DPoints=1)\n"

    def data_attr_name(self):
        return "PointData"

    def color_array_type(self):
        return "POINTS"


class Threshold(PipelineNode):
    """
    This class represents a pipeline node for thresholding.

    You should construct the instance of this class by
    :meth:`PipelineNode.threshold`.
    """

    def __init__(self, parent, field, range):
        super().__init__(parent)
        self.field = field
        self.range = range

    def write_constructor(self, out):
        out.write("{} = Threshold({})\n".format(self.variable(), self.parent.variable()))

    def write_specialized_code(self, out, already_written):
        out.write("{}.Scalars = ['CELLS', r'{}']\n".format(self.variable(), self.field))
        out.write("{}.ThresholdRange = ({}, {})\n".format(self.variable(), self.range[0], self.range[1]))


class SphereClip(PipelineNode):
    """
    This class represents a pipeline node for sphere clipping.

    You should construct the instance of this class
    by :meth:`PipelineNode.clip_sphere`.
    """

    def __init__(self, parent, center, radius, inside_out=True):
        super().__init__(parent)
        self.center = center
        self.radius = radius
        self.inside_out = inside_out

    def write_constructor(self, out):
        out.write(
            "{} = Clip({}, InsideOut={}, ClipType='Sphere')\n".format(
                self.variable(),
                self.parent.variable(),
                int(self.inside_out),
            )
        )

    def write_specialized_code(self, out, already_written):
        out.write("{}.ClipType.Center = [{}, {}, {}]\n".format(self.variable(), *self.center))
        out.write("{}.ClipType.Radius = {}\n".format(self.variable(), self.radius))


def write_python_code(out, sources, bgcolor):
    """
    Write a python code generated by `sources` to `out`.

    Args:
        out (io-like): The output IO object.
        sources (list of PipelineNode): Pipeline nodes to be output.
    """
    already_written = set()
    out.write("from paraview.simple import *\n")
    for source in sources:
        source.write_code(out, already_written, True)
    out.write("view = GetActiveView()\n")
    if bgcolor:
        out.write("view.Background = [{}, {}, {}]\n".format(*bgcolor))
    out.write("Render()\n")


def show(sources, path=None, wait=True, bgcolor=None):
    """
    Shows `sources` by invoking paraview.

    Args:
        sources (list of PipelineNode): Pipeline nodes to be output.
        path (string or None): The output filename. If this parameter
            is None, a temporary filepath is generated and use it.
            If `path` is not None, the file remains after stopping paraview.
        wait (bool): If True, this function returns after paraview stops.
            If False, this function returns immediately after
            paraview is invoked by using backgroup process mechanism.
        bgcolor (None or tuple[float, float, float]): Background color.
    """
    f = open(path, "w") if path else TempFile(".py")
    nodes = [s.to_paraview_node() for s in sources]
    write_python_code(f, nodes, bgcolor)
    f.close()
    # sp.run(["cmd.exe", "/k", "type", path])

    def finalize():
        if isinstance(f, TempFile):
            f.cleanup()
        for node in nodes:
            node.cleanup()

    utils.invoke_paraview("--script={}".format(f.name), wait=wait, finalize=finalize)


current_index = 0


def new_index():
    global current_index
    current_index += 1
    return current_index



import re
import os


def add_arguments_for_load_diagrams(parser):
    parser.add_argument("-d", "--degree", type=int, required=True, help="degree of PH")
    parser.add_argument("-T", "--type", help="input file format (dipha, idipha, text) (default: autodetect)")
    parser.add_argument(
        "-N",
        "--negate",
        default=False,
        action="store_true",
        help="flip the sign of birth/death times for superlevel persistence (default: False)",
    )


def add_arguments_for_histogram_rulers(parser):
    parser.add_argument("-x", "--x-range", type=parse_range, help="birth range")
    parser.add_argument("-X", "--xbins", type=int, default=128, help="number of bins in birth-axis")
    parser.add_argument("-y", "--y-range", type=parse_range, default=None, help="death range")
    parser.add_argument("-Y", "--ybins", type=int, default=None, help="number of bins in death-axis")


def add_arguments_for_gaussian_diffusion(parser):
    parser.add_argument(
        "-D", "--gaussian-sd", type=float, required=True, help="standard deviation of gaussian diffusion"
    )


def add_arguments_for_kernel_weight(parser):
    parser.add_argument("-C", type=float, help="weight constant C")
    parser.add_argument("-p", type=float, default=1.0, help="weight constant p")


FLOAT_REGEXP = r"[+-]?\d+(\.\d+)?"
RANGE_REGEXPS = [
    re.compile(r"\A(?P<begin>{0}):(?P<end>{0})\Z".format(FLOAT_REGEXP)),
    re.compile(r"\A\[(?P<begin>{0}):(?P<end>{0})\]\Z".format(FLOAT_REGEXP)),
]


def parse_range(string):
    for regexp in RANGE_REGEXPS:
        m = regexp.match(string)
        if m:
            return (float(m.group("begin")), float(m.group("end")))
    raise ValueError("{} cannot be parsed as range".format(string))


def parse_bool(string):
    s = string.lower()
    if s == "true" or s == "yes" or s == "1" or s == "on":
        return True
    if s == "false" or s == "no" or s == "0" or s == "off":
        return False
    raise ValueError("{} cannot be parsed as boolean".format(string))


def parse_color(string):
    if re.match(r"#[0-9a-fA-F]{6}\Z", string):
        return tuple(int(string[i : i + 2], 16) for i in [1, 3, 5])
    raise ValueError("{} cannot be parsed as color".format(string))


def parse_cubical_periodic_flags(arg):
    return [bool(int(el)) for el in re.split(",", arg)]


def check_abolished_output(args):
    if os.path.splitext(args.output)[1] in [".complex", ".icomplex", ".diagram", ".idiagram"]:
        raise RuntimeError(
            "Now this program does not support outputting " ".diagram, .complex, .icomplex, and .idiagram files"
        )


def check_cubical_option(args):
    if args.save_boundary_map and not args.cubical:
        raise RuntimeError("-M (--save-boundary-map) option must be with" " -C (--cubical) option")



import argparse
import json

import msgpack

from homcloud.version import __version__
import homcloud.pict.utils as utils
from homcloud.license import add_argument_for_license
import homcloud.pict_tree as ext
from homcloud.visualize_3d import ParaViewLinesDrawer
from .distance_transform import distance_transform
from .binarize_nd import binarize_picture
from homcloud.pdgm_format import PDGMWriter, PDChunk


def argument_parser():
    p = argparse.ArgumentParser(description="Compute 0th PH tree and (n-1)-th PH tree")
    p.add_argument("-V", "--version", action="version", version=__version__)
    add_argument_for_license(p)
    p.add_argument(
        "-m",
        "--mode",
        required=True,
        help=("mode (white-base or black-base for binarize," + " superlevel or sublevel for levelset)"),
    )

    for_binarize = p.add_argument_group("for binarize")
    for_binarize.add_argument("-t", "--threshold", type=float, default=128)
    for_binarize.add_argument("--gt", type=float, default=None, help="lower threshold")
    for_binarize.add_argument("--lt", type=float, default=None, help="upper threshold")
    for_binarize.add_argument("-s", "--ensmall", action="store_true", default=False, help="ensmall binarized picture")
    for_binarize.add_argument(
        "--metric",
        default="manhattan",
        help="metric used to enlarge binarized image" " (manhattan(default), euclidean, etc.)",
    )
    # for_binarize.add_argument("--matrix", help="not implemnted yet")

    # arguments for input and output
    for_input_output = p.add_argument_group("for input and output")
    for_input_output.add_argument(
        "-T",
        "--type",
        default="text_nd",
        help="input data format " "(text2d, text_nd(default), picture2d, picture3d, npy)",
    )
    for_input_output.add_argument(
        "-O", "--output-type", default="pdgm", help="output file format " "(json, msgpack, pdgm(default))"
    )
    for_input_output.add_argument("-o", "--output", required=True, help="output file")
    p.add_argument("input", nargs="*", help="input files")
    return p


def main(args=None):
    args = args or argument_parser().parse_args()
    pict = utils.load_nd_picture(args.input, args.type)
    is_superlevel = args.mode == "superlevel"

    if args.mode in ["black-base", "white-base"]:
        binary_pict = binarize_picture(pict, args.threshold, args.mode, (args.gt, args.lt))
        bitmap = distance_transform(binary_pict, args.metric, args.ensmall)
    elif args.mode in ["sublevel", "superlevel"]:
        bitmap = pict
    else:
        raise RuntimeError("invalid mode")

    lower_mergetree, upper_mergetree = construct_mergetrees(bitmap, is_superlevel)
    dic = construct_dict(bitmap.ndim, is_superlevel, lower_mergetree, upper_mergetree)

    if args.output_type == "json":
        with open(args.output, "w") as f:
            json.dump(dic, f)
    elif args.output_type == "msgpack":
        with open(args.output, "wb") as f:
            msgpack.dump(dic, f, use_bin_type=True)
    elif args.output_type == "pdgm":
        with open(args.output, "wb") as f:
            save_pdgm(f, bitmap.ndim, is_superlevel, lower_mergetree, upper_mergetree)
    else:
        raise RuntimeError("Unknown output format: {}".format(args.output_type))


def save_pdgm(f, dim, sign_flipped, lower_mergetree, upper_mergetree):
    writer = PDGMWriter(f, "bitmap-tree", dim)
    writer.sign_flipped = sign_flipped
    writer.append_chunk(pdchunk(lower_mergetree))
    writer.append_chunk(pdchunk(upper_mergetree))
    writer.append_simple_chunk("bitmap_phtrees", merge_tree_to_dict(lower_mergetree), degree=0)
    writer.append_simple_chunk("bitmap_phtrees", merge_tree_to_dict(upper_mergetree), degree=dim - 1)
    writer.write()


def pdchunk(mt):
    return PDChunk("pd", mt.degree(), *tree_to_pd(mt))


def tree_to_pd(mergetree):
    births = []
    deaths = []
    ess_births = []
    for n in range(mergetree.num_nodes()):
        if mergetree.node_is_trivial(n):
            continue
        if mergetree.node_birth_time(n) == mergetree.node_death_time(n):
            continue
        if mergetree.node_is_essential(n):
            ess_births.append(mergetree.node_birth_time(n))
        else:
            births.append(mergetree.node_birth_time(n))
            deaths.append(mergetree.node_death_time(n))
    return births, deaths, ess_births


def merge_tree_to_dict(mt):
    return {
        "degree": mt.degree(),
        "nodes": {
            mt.node_id(n): {
                "id": mt.node_id(n),
                "birth-time": mt.node_birth_time(n),
                "death-time": mt.node_death_time(n),
                "birth-pixel": mt.node_birth_pixel(n),
                "death-pixel": mt.node_death_pixel(n),
                "volume": mt.node_volume(n),
                "parent": mt.node_parent(n),
                "children": mt.node_children(n),
            }
            for n in range(mt.num_nodes())
            if not mt.node_is_on_boundary(n)
        },
    }


def construct_mergetrees(array, is_superlevel):
    mergetree_lower = ext.MergeTree(array.astype(float), is_superlevel, True)
    mergetree_lower.compute()
    mergetree_upper = ext.MergeTree(array.astype(float), is_superlevel, False)
    mergetree_upper.compute()
    return mergetree_lower, mergetree_upper


def construct_dict(dim, sign_flipped, lower_mergetree, upper_mergetree):
    return {
        "dim": dim,
        "sign-flipped": sign_flipped,
        "lower": merge_tree_to_dict(lower_mergetree),
        "upper": merge_tree_to_dict(upper_mergetree),
    }


def prepare_drawer_for_paraview(n_colors):
    return ParaViewLinesDrawer(n_colors, {"index": None, "isboundary": "1"})


def draw_volume(drawer, volume, color, index):
    processed_pixels = set()
    for pixel in map(tuple, boundary_volume(volume)):
        for adj in adjacent_pixels(pixel):
            if adj in processed_pixels:
                drawer.draw_line(pixel, adj, color, index=index)
        processed_pixels.add(pixel)


def adjacent_pixels(coord):
    coord = list(coord)
    for n in range(len(coord)):
        for s in [-1, 1]:
            coord[n] += s
            yield tuple(coord)
            coord[n] -= s


def boundary_volume(volume):
    def neighbour_pixels(coord):
        coord = list(coord)
        ret = []

        def iter(k):
            if k == len(coord):
                ret.append(tuple(coord))
                return

            for s in [-1, 1]:
                coord[k] += s
                iter(k + 1)
                coord[k] -= s

        iter(0)
        return ret

    volume_set = set(map(tuple, volume))
    result = []
    for pixel in volume:
        if not all(adj in volume_set for adj in neighbour_pixels(pixel)):
            result.append(pixel)
    return result


if __name__ == "__main__":
    main()



import plotly.graph_objects as go
import numpy as np
import itertools


def PointCloud(array, color=None, size=1, name=""):
    """
    Returns a plotly's trace object for a pointcloud.

    Args:
        array (np.array): 3D pointcloud
        color (string or None): The name of color
        size (int): The size of each point
        name (string): The name of pointcloud

    Returns:
        plotly.graph_objects.Scatter3d: Plotly's trace object for the visualization of the pointcloud
    """
    assert array.ndim == 2 and array.shape[1] >= 3
    if array.shape[1] == 4 and color is True:
        color = array[:, 3]
    return go.Scatter3d(
        x=array[:, 0], y=array[:, 1], z=array[:, 2], name=name, mode="markers", marker=dict(size=size, color=color)
    )


def SimplicesMesh(simplices, color=None, name=""):
    vertices = dict()
    coordinates = []
    n = 0
    for simplex in simplices:
        for vertex in simplex:
            v = tuple(vertex)
            if v not in vertices:
                vertices[v] = n
                coordinates.append(vertex)
                n += 1

    simplices_by_vertex_index = np.array([[vertices[tuple(v)] for v in simplex] for simplex in simplices])
    coordinates = np.array(coordinates)

    return go.Mesh3d(
        x=coordinates[:, 0],
        y=coordinates[:, 1],
        z=coordinates[:, 2],
        i=simplices_by_vertex_index[:, 0],
        j=simplices_by_vertex_index[:, 1],
        k=simplices_by_vertex_index[:, 2],
        name=name,
        color=color,
    )


def Simplices(simplices, color=None, width=1, name=""):
    d = len(simplices[0])
    ij_pairs = [(i, j) for j in range(d) for i in range(j + 1, d)]
    xs = []
    ys = []
    zs = []
    for simplex in simplices:
        for i, j in ij_pairs:
            xs.extend([simplex[i][0], simplex[j][0], None])
            ys.extend([simplex[i][1], simplex[j][1], None])
            zs.extend([simplex[i][2], simplex[j][2], None])
    return go.Scatter3d(
        x=xs,
        y=ys,
        z=zs,
        mode="lines",
        name=name,
        showlegend=False,
        line=dict(color=color, width=width),
    )


def Loop(vertices, color=None, width=1, name=""):
    vs = np.concatenate([vertices, [vertices[0]]])
    return go.Scatter3d(
        x=vs[:, 0],
        y=vs[:, 1],
        z=vs[:, 2],
        mode="lines",
        line=dict(color=color, width=width),
        name=name,
        showlegend=False,
    )


class VertexIndexer(object):
    def __init__(self):
        self.vertex2index = dict()
        self.index2vertex = list()
        self.last_index = 0

    def index_of(self, vertex):
        if vertex in self.vertex2index:
            return self.vertex2index[vertex]
        self.vertex2index[vertex] = self.last_index
        self.index2vertex.append(vertex)
        self.last_index += 1

        return self.last_index - 1

    def coordinates(self):
        return np.array(self.index2vertex)


class VoxelDrawer(object):
    def __init__(self, bitmap, color, name, offsets):
        self.bitmap = bitmap
        self.shape = bitmap.shape
        self.color = color
        self.offsets = offsets
        self.vertex_indexer = VertexIndexer()
        self.triangles = []
        self.name = name

    def isface(self, x, y, z, dx, dy, dz):
        return self.at(x, y, z) != self.at(x + dx, y + dy, z + dz)

    def at(self, x, y, z):
        if x < 0 or y < 0 or z < 0:
            return False
        if x >= self.shape[2] or y >= self.shape[1] or z >= self.shape[0]:
            return False
        return self.bitmap[z, y, x]

    def putface(self, x, y, z, dx, dy, dz):
        vertices = self.square_vertices(x, y, z, dx, dy, dz)
        vertex_indices = [self.vertex_indexer.index_of(vertex) for vertex in vertices]
        self.triangles.append([vertex_indices[0], vertex_indices[1], vertex_indices[2]])
        self.triangles.append([vertex_indices[1], vertex_indices[3], vertex_indices[2]])

    def square_vertices(self, x, y, z, dx, dy, dz):
        z += self.offsets[0]
        y += self.offsets[1]
        x += self.offsets[2]

        if dx == 1:
            return (
                (x + 0.5, y - 0.5, z - 0.5),
                (x + 0.5, y + 0.5, z - 0.5),
                (x + 0.5, y - 0.5, z + 0.5),
                (x + 0.5, y + 0.5, z + 0.5),
            )
        if dy == 1:
            return (
                (x - 0.5, y + 0.5, z - 0.5),
                (x + 0.5, y + 0.5, z - 0.5),
                (x - 0.5, y + 0.5, z + 0.5),
                (x + 0.5, y + 0.5, z + 0.5),
            )
        if dz == 1:
            return (
                (x - 0.5, y - 0.5, z + 0.5),
                (x + 0.5, y - 0.5, z + 0.5),
                (x - 0.5, y + 0.5, z + 0.5),
                (x + 0.5, y + 0.5, z + 0.5),
            )

    def graph_object(self):
        for dx, dy, dz in [(1, 0, 0), (0, 1, 0), (0, 0, 1)]:
            for z in range(-1, self.shape[0] + 1):
                for y in range(-1, self.shape[1] + 1):
                    for x in range(-1, self.shape[2] + 1):
                        if self.isface(x, y, z, dx, dy, dz):
                            self.putface(x, y, z, dx, dy, dz)

        coordinates = self.vertex_indexer.coordinates()
        triangles = np.array(self.triangles)
        return go.Mesh3d(
            x=coordinates[:, 0],
            y=coordinates[:, 1],
            z=coordinates[:, 2],
            i=triangles[:, 0],
            j=triangles[:, 1],
            k=triangles[:, 2],
            name=self.name,
            color=self.color,
        )


def Bitmap3d(bitmap, color=None, name="", offsets=[0, 0, 0]):
    """
    Returns a plotly's trace object for a 3D bitmap.

    Args:
        bitmap (np.array): 3D boolean bitmap
        color (string or None): The name of color
        name (string): The name of pointcloud

    Returns:
        plotly.graph_objects.Mesh3d: Plotly's trace object for the visualization of the pointcloud
    """
    return VoxelDrawer(bitmap, color, name, offsets).graph_object()


def Voxels(voxels, color=None, name=""):
    assert len(voxels[0]) == 3
    zmax, ymax, xmax = np.max(voxels, axis=0) + 1
    bitmap = np.zeros((zmax, ymax, xmax), bool)
    for v in voxels:
        bitmap[v[0], v[1], v[2]] = True
    return Bitmap3d(bitmap, color, name)


def Cubes1d(cubes, color, width, name):
    zs = []
    ys = []
    xs = []
    for cube, d in cubes:
        zs.extend([cube[0], cube[0] + d[0], None])
        ys.extend([cube[1], cube[1] + d[1], None])
        xs.extend([cube[2], cube[2] + d[2], None])
    return go.Scatter3d(
        x=xs,
        y=ys,
        z=zs,
        mode="lines",
        name=name,
        showlegend=False,
        line=dict(color=color, width=width),
    )


def rotate3(l, k):
    return l[k % 3], l[(k + 1) % 3], l[(k + 2) % 3]


def ndeg2rot(ndeg):
    return ndeg.index(0)


def Cubes2d(cubes, color, width, name):
    xyzs = [[], [], []]
    for cube, d in cubes:
        cs1, cs2, cs3 = rotate3(xyzs, ndeg2rot(d))
        c1, c2, c3 = rotate3(cube, ndeg2rot(d))
        cs1.extend([c1, c1, c1, c1, c1, None])
        cs2.extend([c2, c2 + 1, c2 + 1, c2, c2, None])
        cs3.extend([c3, c3, c3 + 1, c3 + 1, c3, None])

    return go.Scatter3d(
        x=xyzs[0],
        y=xyzs[1],
        z=xyzs[2],
        mode="lines",
        name=name,
        showlegend=False,
        line=dict(color=color, width=width),
    )


def Cubes(cubes, color=None, width=1, name=""):
    dim = sum(cubes[0][1])
    if dim == 1:
        return Cubes1d(cubes, color, width, name)
    elif dim == 2:
        return Cubes2d(cubes, color, width, name)
    elif dim == 3:
        raise RuntimeError("Dim 3 cube is not visualizable by homcloud-plotly")


class CubeMeshDrawer(object):
    def __init__(self, cubes, color, name):
        self.cubes = cubes
        self.color = color
        self.name = name
        self.vertex_indexer = VertexIndexer()

    UV = [
        (np.array([0, 1, 0]), np.array([0, 0, 1])),
        (np.array([0, 0, 1]), np.array([1, 0, 0])),
        (np.array([1, 0, 0]), np.array([0, 1, 0])),
    ]

    def graph_object(self):
        triangles = []

        for cube, ndeg in self.cubes:
            u, v = self.UV[ndeg.index(0)]
            p0 = self.vertex_indexer.index_of(tuple(cube))
            p1 = self.vertex_indexer.index_of(tuple(u + cube))
            p2 = self.vertex_indexer.index_of(tuple(u + v + cube))
            p3 = self.vertex_indexer.index_of(tuple(v + cube))
            triangles.append([p0, p1, p2])
            triangles.append([p0, p2, p3])

        coordinates = self.vertex_indexer.coordinates()
        triangles = np.array(triangles)
        return go.Mesh3d(
            x=coordinates[:, 0],
            y=coordinates[:, 1],
            z=coordinates[:, 2],
            i=triangles[:, 0],
            j=triangles[:, 1],
            k=triangles[:, 2],
            name=self.name,
            color=self.color,
        )


def CubesMesh(cubes, color=None, name=""):
    dim = sum(cubes[0][1])
    assert dim == 2

    return CubeMeshDrawer(cubes, color, name).graph_object()


def SimpleScene(xrange=None, yrange=None, zrange=None):
    """
    Returns a plotly's `graph_object.layout.Scene` object to remove x/y/z axes.

    You can use `xrange`/`yrange`/`zrange` to cut off the visualization area

    Args:
        xrange ((float, float) or None): The range of X axis to be visualized
        yrange ((float, float) or None): The range of Y axis to be visualized
        zrange ((float, float) or None): The range of Z axis to be visualized

    Returns:
        plotly.graph_objects.layout.Scene: Plotly's scene object
    """
    return go.layout.Scene(
        xaxis=dict(range=xrange, visible=False),
        yaxis=dict(range=yrange, visible=False),
        zaxis=dict(range=zrange, visible=False),
    )



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui/view_index.ui'
#
# Created by: PyQt5 UI code generator 5.15.6
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(800, 600)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.centralwidget)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.scrollArea = QtWidgets.QScrollArea(self.centralwidget)
        self.scrollArea.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 374, 508))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.scrollAreaWidgetContents)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label_Pict = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_Pict.setText("")
        self.label_Pict.setObjectName("label_Pict")
        self.horizontalLayout_2.addWidget(self.label_Pict)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.horizontalLayout.addWidget(self.scrollArea)
        self.frame_2 = QtWidgets.QFrame(self.centralwidget)
        self.frame_2.setMaximumSize(QtCore.QSize(500, 16777215))
        self.frame_2.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame_2.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_2.setObjectName("frame_2")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.frame_2)
        self.verticalLayout.setObjectName("verticalLayout")
        self.listView_Pairs = QtWidgets.QListView(self.frame_2)
        self.listView_Pairs.setObjectName("listView_Pairs")
        self.verticalLayout.addWidget(self.listView_Pairs)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.label = QtWidgets.QLabel(self.frame_2)
        self.label.setMaximumSize(QtCore.QSize(100, 16777215))
        self.label.setObjectName("label")
        self.horizontalLayout_3.addWidget(self.label)
        self.spinBox_Scale = QtWidgets.QSpinBox(self.frame_2)
        self.spinBox_Scale.setMinimum(1)
        self.spinBox_Scale.setMaximum(10)
        self.spinBox_Scale.setObjectName("spinBox_Scale")
        self.horizontalLayout_3.addWidget(self.spinBox_Scale)
        self.verticalLayout.addLayout(self.horizontalLayout_3)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.label_2 = QtWidgets.QLabel(self.frame_2)
        self.label_2.setObjectName("label_2")
        self.horizontalLayout_4.addWidget(self.label_2)
        self.edit_MinBirth = QtWidgets.QLineEdit(self.frame_2)
        self.edit_MinBirth.setObjectName("edit_MinBirth")
        self.horizontalLayout_4.addWidget(self.edit_MinBirth)
        self.verticalLayout.addLayout(self.horizontalLayout_4)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.label_3 = QtWidgets.QLabel(self.frame_2)
        self.label_3.setObjectName("label_3")
        self.horizontalLayout_5.addWidget(self.label_3)
        self.edit_MaxBirth = QtWidgets.QLineEdit(self.frame_2)
        self.edit_MaxBirth.setObjectName("edit_MaxBirth")
        self.horizontalLayout_5.addWidget(self.edit_MaxBirth)
        self.verticalLayout.addLayout(self.horizontalLayout_5)
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.label_4 = QtWidgets.QLabel(self.frame_2)
        self.label_4.setObjectName("label_4")
        self.horizontalLayout_6.addWidget(self.label_4)
        self.edit_MinDeath = QtWidgets.QLineEdit(self.frame_2)
        self.edit_MinDeath.setObjectName("edit_MinDeath")
        self.horizontalLayout_6.addWidget(self.edit_MinDeath)
        self.verticalLayout.addLayout(self.horizontalLayout_6)
        self.horizontalLayout_7 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_7.setObjectName("horizontalLayout_7")
        self.label_5 = QtWidgets.QLabel(self.frame_2)
        self.label_5.setObjectName("label_5")
        self.horizontalLayout_7.addWidget(self.label_5)
        self.edit_MaxDeath = QtWidgets.QLineEdit(self.frame_2)
        self.edit_MaxDeath.setObjectName("edit_MaxDeath")
        self.horizontalLayout_7.addWidget(self.edit_MaxDeath)
        self.verticalLayout.addLayout(self.horizontalLayout_7)
        self.horizontalLayout_8 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_8.setObjectName("horizontalLayout_8")
        self.label_6 = QtWidgets.QLabel(self.frame_2)
        self.label_6.setObjectName("label_6")
        self.horizontalLayout_8.addWidget(self.label_6)
        self.edit_MinLifetime = QtWidgets.QLineEdit(self.frame_2)
        self.edit_MinLifetime.setObjectName("edit_MinLifetime")
        self.horizontalLayout_8.addWidget(self.edit_MinLifetime)
        self.verticalLayout.addLayout(self.horizontalLayout_8)
        self.horizontalLayout_9 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_9.setObjectName("horizontalLayout_9")
        self.label_7 = QtWidgets.QLabel(self.frame_2)
        self.label_7.setObjectName("label_7")
        self.horizontalLayout_9.addWidget(self.label_7)
        self.edit_MaxLifetime = QtWidgets.QLineEdit(self.frame_2)
        self.edit_MaxLifetime.setObjectName("edit_MaxLifetime")
        self.horizontalLayout_9.addWidget(self.edit_MaxLifetime)
        self.verticalLayout.addLayout(self.horizontalLayout_9)
        self.button_Filter = QtWidgets.QPushButton(self.frame_2)
        self.button_Filter.setObjectName("button_Filter")
        self.verticalLayout.addWidget(self.button_Filter)
        self.horizontalLayout.addWidget(self.frame_2)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 800, 35))
        self.menubar.setObjectName("menubar")
        self.menuFile = QtWidgets.QMenu(self.menubar)
        self.menuFile.setObjectName("menuFile")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.actionQuit = QtWidgets.QAction(MainWindow)
        self.actionQuit.setObjectName("actionQuit")
        self.menuFile.addAction(self.actionQuit)
        self.menubar.addAction(self.menuFile.menuAction())

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "View index for a picture"))
        self.label.setText(_translate("MainWindow", "Zoom:"))
        self.label_2.setText(_translate("MainWindow", "min birth:"))
        self.label_3.setText(_translate("MainWindow", "max birth:"))
        self.label_4.setText(_translate("MainWindow", "min death:"))
        self.label_5.setText(_translate("MainWindow", "max death:"))
        self.label_6.setText(_translate("MainWindow", "min lifetime:"))
        self.label_7.setText(_translate("MainWindow", "max lifetime:"))
        self.button_Filter.setText(_translate("MainWindow", "Apply Filter"))
        self.menuFile.setTitle(_translate("MainWindow", "File"))
        self.actionQuit.setText(_translate("MainWindow", "Quit"))
        self.actionQuit.setShortcut(_translate("MainWindow", "Ctrl+Q"))



class HomCloudError(Exception):
    """
    Base exception class of all exceptions in homcloud.interface
    and homcloud.paraview_interface module.
    """

    def __init__(self, message, code):
        self.message = message
        self.code = code


class VolumeNotFound(HomCloudError):
    """
    Exception class for :meth:`Pair.optimal_volume` and
    :meth:`Pair.stable_volume`.
    """

    def __init__(self, message, code):
        super().__init__(message, code)



import argparse
import sys

from homcloud.version import __version__
from homcloud.argparse_common import add_arguments_for_load_diagrams, parse_bool
import homcloud.pdgm_format as pdgm_format
from homcloud.pdgm import PDGM


def main(args=None):
    args = args or argument_parser().parse_args()
    output_text(args, get_output_io(args))


def argument_parser():
    """Returns the parser of args"""
    p = argparse.ArgumentParser(description="Dump pairs in dipha's diagram")
    p.add_argument("-V", "--version", action="version", version=__version__)
    add_arguments_for_load_diagrams(p)
    p.add_argument("-o", "--output", help="output text file")
    p.add_argument(
        "-S",
        "--show-simplices",
        type=parse_bool,
        default=False,
        help="show birth/death simplices (yes/no, default:no)",
    )
    p.add_argument(
        "-E",
        "--show-essential-pairs",
        type=parse_bool,
        default=False,
        help="show essential pairs (yes/no, default:no)",
    )
    p.add_argument(
        "-s",
        "--symbols",
        type=parse_bool,
        default=False,
        help="show birth/death simplices by symbols" "(yes/no default:no)",
    )
    p.add_argument("input", help="input pdgm file")
    return p


def get_output_io(args):
    if args.output:
        return open(args.output, "w")
    else:
        return sys.stdout


def output_text(args, output):
    if args.show_simplices:
        pdgm = PDGM.open(args.input, args.degree)
        if pdgm.filtration_type == "bitmap":
            resolve_index = pdgm.index_to_pixel.__getitem__
        else:
            resolve_index = pdgm.geometry_resolver(args.symbols).resolve_cell
        write_pairs_with_geometry(pdgm, output, args.show_essential_pairs, resolve_index)
    else:
        write_pairs(PDGM.open(args.input, args.degree, False), output, args.show_essential_pairs)


def write_pairs(diagram, io, show_essential_pairs):
    for birth, death in zip(diagram.births, diagram.deaths):
        io.write("{} {}\n".format(birth, death))

    if show_essential_pairs:
        inf = inf_string(diagram)
        for birth in diagram.essential_births:
            io.write("{} {}\n".format(birth, inf))


def write_pairs_with_geometry(pdgm, io, show_essential_pairs, resolve_index):
    def format_geometry(index):
        def format(obj):
            if isinstance(obj, list):
                return "(" + ",".join(format(item) for item in obj) + ")"
            else:
                return str(obj)

        return format(resolve_index(index))

    pairs = zip(pdgm.births, pdgm.deaths, pdgm.birth_indices, pdgm.death_indices)
    for birth, death, birth_index, death_index in pairs:
        io.write("{} {} {} {}\n".format(birth, death, format_geometry(birth_index), format_geometry(death_index)))

    if show_essential_pairs:
        inf = inf_string(pdgm)
        for birth, birth_index in zip(pdgm.essential_births, pdgm.essential_birth_indices):
            io.write("{} {} {}\n".format(birth, inf, format_geometry(birth_index)))


def inf_string(diagram):
    return "-inf" if diagram.sign_flipped else "inf"


if __name__ == "__main__":
    main()



import argparse
from tempfile import TemporaryDirectory
import os
import subprocess
import re

from PIL import Image
import numpy as np


def main(args=None):
    args = args or argument_parser().parse_args()
    volumes = [np.load(path) for path in args.input]

    assert [volume.shape == volumes[0].shape for volume in volumes]

    if args.slice is not None:
        with TemporaryDirectory() as tmpdir:
            write_volume_slices(volumes, args.slice, args.spacer, args.range, tmpdir)
            subprocess.call("{} {}".format(args.image_viewer, tmpdir), shell=True)


def write_volume_slices(volumes, direction, spacer, slice_range, dirname):
    uppers = list(map(np.max, volumes))
    lowers = list(map(np.min, volumes))
    shape = volumes[0].shape[0:direction] + volumes[0].shape[direction + 1 :]
    width = shape[1] * len(volumes) + spacer * (len(volumes) - 1)
    height = shape[0]
    d = shape[1] + spacer
    if not slice_range:
        slice_range = (0, volumes[0].shape[direction])
    for n in range(*slice_range):
        array = np.zeros((height, width), dtype=np.uint8)
        for i, (volume, u, l) in enumerate(zip(volumes, uppers, lowers)):
            u_f = float(u)
            l_f = float(l)
            slice = slice_volume(volume, n, direction)
            array[:, d * i : d * i + shape[1]] = (slice - l_f) * 254 / (u_f - l_f)
        path = os.path.join(dirname, "{:04d}.png".format(n))
        Image.fromarray(array).save(path)


def slice_volume(volume, n, direction):
    if direction == 0:
        return volume[n, :, :]
    if direction == 1:
        return volume[:, n, :]
    if direction == 2:
        return volume[:, :, n]
    raise ValueError("{} is not valid for direction".format(n))


def argument_parser():
    p = argparse.ArgumentParser(description="3d npy data viewer")
    p.add_argument("-s", "--slice", default=0, type=int, help="slicing direction (0, 1, or 2)")
    p.add_argument("-S", "--spacer", default=10, type=int, help="spacer pixels")
    p.add_argument("-r", "--range", default=None, type=parse_range, help="range of slices")
    p.add_argument("--image-viewer", default="eog -n", help="image viewer program name")
    p.add_argument("input", nargs="+", help="input files")
    return p


def parse_range(string):
    l, r = re.split(r":", string)
    return int(l), int(r)


if __name__ == "__main__":
    main()



import collections

import numpy as np

from homcloud.visualize_3d import ParaViewLinesDrawer


def search(bitmap, birth_level, start):
    ndim = bitmap.ndim
    shape = bitmap.shape
    queue = collections.deque()
    memo = np.full_like(bitmap, None, dtype=object)

    def backward_list(at):
        lst = [at]
        while memo[at]:
            _, next_pos = memo[at]
            lst.append(next_pos)
            at = next_pos
        return lst

    def next_from(at):
        for i in range(ndim):
            for sign in [-1, +1]:
                n = list(at)
                n[i] += sign
                next_pos = tuple(n)
                if next_pos[i] < 0 or next_pos[i] >= shape[i]:
                    continue
                if bitmap[next_pos] < birth_level:
                    yield next_pos

    def step_from(head_name, at):
        for next_pos in next_from(at):
            next_step = memo[next_pos]
            if next_step:
                next_head_name, next_backward = next_step
                if next_head_name == head_name:
                    continue
                else:
                    return list(reversed(backward_list(at))) + backward_list(next_pos)
            else:
                memo[next_pos] = (head_name, at)
                queue.append((head_name, next_pos))

    def initialize():
        for head_name, pos in enumerate(next_from(start)):
            memo[pos] = (head_name, start)
            queue.append((head_name, pos))

    initialize()

    while True:
        if not len(queue):
            return None
        head_name, at = queue.popleft()
        result = step_from(head_name, at)
        if result:
            return result


class Finder(object):
    def __init__(self, pdgm):
        assert pdgm.filtration_type == "bitmap"
        self.pdgm = pdgm

    def query_pair(self, birth, death):
        path = search(self.bitmap, birth, self.start_pixel(birth))
        return Result(self.pdgm.index_to_level[birth], self.pdgm.index_to_level[death], path)

    @property
    def bitmap(self):
        return self.pdgm.indexed_bitmap

    def start_pixel(self, birth):
        return tuple(self.pdgm.index_to_pixel[birth])


class Result(object):
    def __init__(self, birth_time, death_time, path):
        self.birth_time = birth_time
        self.death_time = death_time
        self.path = path

    def to_jsondict(self):
        return {
            "birth-time": self.birth_time,
            "death-time": self.death_time,
            "path": list(map(list, self.path)),
            "boundary-points": list(map(list, self.boundary_points())),
        }

    def boundary_points(self):
        return set(self.path)

    def draw(self, drawer, color, index):
        drawer.draw_loop(self.path, color, index=index)


def prepare_drawer_for_paraview(n_colors):
    return ParaViewLinesDrawer(n_colors, {"index": None, "isboundary": "1"})



import pyvista as pv
import numpy as np


def Bitmap3D(array):
    """
    Returns PyVista's mesh object for 3d array bitmap.

    Args:
        array (np.array): 3d array

    Returns:
        pyvista.UniformGrid: PyVista's grid object

    """
    grids = pv.UniformGrid(dimensions=np.array(array.shape) + 1, origin=(-0.5, -0.5, -0.5))
    grids.cell_data["values"] = array.flatten(order="F")
    return grids


def Lines(lines):
    lines = np.array(lines)
    assert lines.ndim == 3 and lines.shape[1] == 2 and lines.shape[2] == 3
    return pv.PolyData(
        lines.reshape((lines.shape[0] * 2, 3)), lines=np.array([[2, 2 * i, 2 * i + 1] for i in range(lines.shape[0])])
    )


def Loop(points):
    points = np.array(points)
    assert points.ndim == 2 and points.shape[1] == 3
    return pv.PolyData(points, lines=[[2, k, (k + 1) % points.shape[0]] for k in range(points.shape[0])])


def Triangles(triangles):
    triangles = np.array(triangles)
    assert triangles.shape[1] == 3 and triangles.shape[2] == 3
    return pv.PolyData(
        triangles.reshape((triangles.shape[0] * 3, 3)),
        faces=[[3, k * 3, k * 3 + 1, k * 3 + 2] for k in range(triangles.shape[0])],
        n_faces=triangles.shape[0],
    )


D = np.array(
    [
        [-0.5, -0.5, -0.5],
        [+0.5, -0.5, -0.5],
        [+0.5, +0.5, -0.5],
        [-0.5, +0.5, -0.5],
        [-0.5, -0.5, +0.5],
        [+0.5, -0.5, +0.5],
        [+0.5, +0.5, +0.5],
        [-0.5, +0.5, +0.5],
    ]
)


def SparseVoxels(coords):
    coords = np.array(coords)
    assert coords.ndim == 2 and coords.shape[1] == 3
    n = coords.shape[0]
    return pv.UnstructuredGrid(
        {pv.CellType.HEXAHEDRON: np.arange(n * 8).reshape(n, 8)},
        (coords.reshape(n, 1, 3) + D.reshape(1, 8, 3)).reshape(n * 8, 3),
    )



import numpy as np
from scipy.ndimage import gaussian_filter
from forwardable import forwardable


@forwardable()
class Histogram(object):
    """A histogram, normally used for representing persistence diagrams.
    The PDHistogram subclass is used for this purpose.

    This class is also used for representing vectorized histogram.
    """

    def __init__(self, values, histospec, ess_values=[]):
        self.values = values
        self.histospec = histospec
        self.values_as_float = values.astype(float)
        self.ess_values = ess_values

    def_delegators("histospec", "xedges, yedges, sign_flipped")
    def_delegators("histospec", "xy_extent, x_range, y_range, x_bins, y_bins")

    def maxvalue(self):
        return max([np.max(self.values), np.max(np.array(self.ess_values, dtype=float), initial=-np.inf)])

    def vectorize(self):
        """Return vectorized histogram as a 1D ndarray of floats."""
        return self.values[self.histospec.vectorize_mask()]

    def value_at(self, birth, death):
        """Internally used by vectorize()."""
        y_index = np.searchsorted(self.yedges, death)
        x_index = np.searchsorted(self.xedges, birth)
        if not ((0 < y_index < self.yedges.size) and (0 < x_index < self.xedges.size)):
            return None
        return self.values[y_index - 1, x_index - 1]

    @classmethod
    def reconstruct_from_vector(cls, vector, histinfo):
        """Construct a Histogram object from vector and histinfo

        Args:
        vector -- 1D ndarray of floats, this vector is translated into a histogram.
        histinfo -- a dict that has histogram information.
          The information is generated in vectorize_PD.save_histogram_information().
        """
        x_edges = np.array(histinfo["x-edges"], dtype=float)
        y_edges = np.array(histinfo["y-edges"], dtype=float)
        y_indices = histinfo["y-indices"]
        x_indices = histinfo["x-indices"]
        values = np.zeros((len(y_edges) - 1, len(x_edges) - 1), dtype=vector.dtype)
        for y, x, val in zip(y_indices, x_indices, vector):
            values[y, x] = val

        return cls(values, HistoSpec(x_edges, y_edges))

    def binary_histogram_by_ranking(self, rank_range, sign=None, use_abs=True):
        """Create a binary (bool) histogram by ranking of histogram bin values.

        Args:
        rank_range: a pair of (minimum rank, maximum rank)
        sign: if this value is not None, only bins with positive(sign=+1) or
              negative(sign=-1) value is selected
        use_abs: if true, to compute a ranking, absolute value of each bin is used
        """
        if use_abs:
            values = np.abs(self.values).flatten()
        else:
            values = self.values.flatten()

        min_rank, max_rank = rank_range
        ranks = np.argsort(values.flatten())[::-1][min_rank : max_rank - min_rank + 1]
        mask = np.full_like(values, False, dtype=bool)
        mask[ranks] = True

        mask = mask.reshape(self.values.shape)
        if sign is not None:
            mask &= sign * self.values > 0
        return BinaryHistogram(mask, self.xedges, self.yedges)

    def apply_gaussian_filter(self, sigma):
        """Apply a Gaussian filter to the histogram.

        Note: This method works correctly only if the scale of x-axis and
        the scale of y-axis are the same. This should be fixed in the future.

        Args:
        sigma -- standard deviation of the gaussian distribution
        """
        self.values = gaussian_filter(self.values, self.histospec.normalized_standard_deviation(sigma))

    def apply_weight(self, weight_func):
        """Apply (multiply) weights to all values in bins.

        The weight is computed by weight_func.
        Args:
        weight_func -- an applicable object with two parameters,
          x (birth time) and y (death time)
        """
        xs, ys = self.histospec.coordinates_of_center_of_bins()
        self.values *= np.vectorize(weight_func)(xs, ys)

    def multiply_histogram(self, val):
        """Apply (multiply) a uniform weight to all values in bins.

        Args:
        val -- weight
        """
        self.values *= val


class Ruler(object):
    """A pair of min-max (range) and # of bins."""

    def __init__(self, minmax, bins):
        self.minmax = minmax
        self.bins = bins

    def min(self):
        return self.minmax[0]

    def max(self):
        return self.minmax[1]

    def bin_width(self):
        return (self.max() - self.min()) / float(self.bins)

    def edges(self):
        return np.linspace(self.minmax[0], self.minmax[1], self.bins + 1)

    @staticmethod
    def create_xy_rulers(x_range, xbins, y_range, ybins, diagram):
        x_range = x_range or diagram.minmax_of_birthdeath_time()
        y_range = y_range or x_range
        ybins = ybins or xbins
        return (Ruler(x_range, xbins), Ruler(y_range, ybins))


class HistoSpec(object):
    def __init__(self, xedges, yedges, sign_flipped=False, x_ruler=None, y_ruler=None):
        self.xedges = xedges
        self.yedges = yedges
        self.sign_flipped = sign_flipped
        self.x_ruler = x_ruler
        self.y_ruler = y_ruler

    @staticmethod
    def from_rulers(x_ruler, y_ruler, sign_flipped=False):
        return HistoSpec(x_ruler.edges(), y_ruler.edges(), sign_flipped, x_ruler, y_ruler)

    def xy_extent(self):
        """Return [xmin, xmax, ymin, ymax]."""
        return [self.xedges[0], self.xedges[-1], self.yedges[0], self.yedges[-1]]

    def x_range(self):
        """Return the range of the histogram along with x-axis as a pair of (min, max)."""
        return (self.xedges[0], self.xedges[-1])

    def y_range(self):
        """Return the range of the histogram along with y-axis as a pair of (min, max)."""
        return (self.yedges[0], self.yedges[-1])

    def x_bins(self):
        """Return the number of bins of x-axis."""
        return self.xedges.size - 1

    def y_bins(self):
        """Return the number of bins of y-axis."""
        return self.yedges.size - 1

    def vectorize_mask(self):
        """Internally used by vectorize()."""
        if self.sign_flipped:
            return self.xedges_of_bins() > self.yedges_of_bins()
        else:
            return self.xedges_of_bins() < self.yedges_of_bins()

    def xedges_of_bins(self):
        """Internally used by vectorize()."""
        xedges = self.xedges[1:] if self.sign_flipped else self.xedges[:-1]
        return np.repeat(xedges.reshape((1, self.x_bins())), self.y_bins(), axis=0)

    def yedges_of_bins(self):
        """Internally used by vectorize()."""
        yedges = self.yedges[:-1] if self.sign_flipped else self.yedges[1:]
        return np.repeat(yedges.reshape((self.y_bins(), 1)), self.x_bins(), axis=1)

    def normalized_standard_deviation(self, sigma):
        """Compute the standard deviation from x-scale and sigma.

        This method is internally used from apply_gaussian_filter.
        """
        return sigma / self.x_ruler.bin_width()

    def coordinates_of_center_of_bins(self):
        """Internally used by vectorize()."""
        x_centers = (self.xedges[:-1] + self.xedges[1:]) / 2
        y_centers = (self.yedges[:-1] + self.yedges[1:]) / 2
        xs = np.repeat(x_centers.reshape((1, self.x_bins())), self.y_bins(), axis=0)
        ys = np.repeat(y_centers.reshape((self.y_bins(), 1)), self.x_bins(), axis=1)
        return (xs, ys)

    def centers_of_vectorize_bins(self):
        xs, ys = self.coordinates_of_center_of_bins()
        mask = self.vectorize_mask()
        return np.vstack((xs[mask], ys[mask])).transpose()

    def indices_for_vectorization(self):
        """Internally used by vectorize()."""
        yindices, xindices = np.indices((self.y_bins(), self.x_bins()))
        mask = self.vectorize_mask()
        return (yindices[mask], xindices[mask])

    def vector_size(self):
        return np.sum(self.vectorize_mask())

    def histogram_from_diagram(self, diagram, cls=Histogram):
        values, *_ = np.histogram2d(
            diagram.deaths,
            diagram.births,
            bins=(self.y_bins(), self.x_bins()),
            range=((self.y_range(), self.x_range())),
        )
        ess_values, _ = np.histogram(diagram.essential_births, bins=self.x_bins(), range=self.x_range())

        return cls(values, self, ess_values)

    def histogram_from_vector(self, vector, cls=Histogram):
        y_indices, x_indices = self.indices_for_vectorization()
        values = np.zeros((self.y_bins(), self.x_bins()), dtype=vector.dtype)
        for y, x, val in zip(y_indices, x_indices, vector):
            values[y, x] = val

        return cls(values, self)


def PDHistogram(diagram, x_ruler, y_ruler):
    spec = HistoSpec.from_rulers(x_ruler, y_ruler, diagram.sign_flipped)
    return spec.histogram_from_diagram(diagram)


class BinaryHistogram(Histogram):
    """A PD histogram whose values are bool."""

    def filter_pairs(self, pairs):
        return [pair for pair in pairs if self.value_at(pair.birth, pair.death)]



import math
import re

import msgpack
import numpy as np

import homcloud.dipha as dipha
import homcloud.homccube as homccube


class Filtration(object):
    """Super class of AlphaFiltration, BitmapFiltration and CubicalFiltration"""

    def compute_pdgm_and_save(self, path, algorithm=None, save_suppl_info=True):
        with open(path, "wb") as f:
            self.compute_pdgm(f, algorithm, save_suppl_info)

    def compute_pairs(self, algorithm=None, parallels=1, upper_dim=None, upper_value=None):
        def compute_pairs_by_phat():
            matrix = self.build_phat_matrix()
            matrix.reduce(algorithm)
            return matrix.birth_death_pairs(), matrix.boundary_map_byte_sequence()

        def format_pair(pair):
            d, birth, death = pair
            return (d, int(birth), (None if death == math.inf else int(death)))

        def compute_pairs_by_dipha():
            return [format_pair(pair) for pair in dipha.pairs_from_filtration(self)]

        def compute_pairs_by_homccube():
            return homccube.compute_pd(
                self.index_bitmap.astype(np.int32), self.periodicity, self.parse_homccube_algorithm(algorithm)
            )

        algorithm = algorithm or self.favorite_algorithm()

        if algorithm.startswith("phat-"):
            return compute_pairs_by_phat()
        if algorithm.startswith("dipha"):
            return compute_pairs_by_dipha(), None
        if algorithm.startswith("homccube-"):
            return compute_pairs_by_homccube(), None

        raise NotImplementedError("unknown alogrithm: {}".format(algorithm))

    @staticmethod
    def parse_homccube_algorithm(algorithm):
        return int(re.match(r"homccube-(\d+)", algorithm).group(1))



import argparse

import numpy as np

import homcloud.pict.utils as utils
from homcloud.version import __version__
from homcloud.license import add_argument_for_license


def argument_parser():
    p = argparse.ArgumentParser(description="Convert data to an .npy file")
    p.add_argument("-V", "--version", action="version", version=__version__)
    p.add_argument(
        "-T",
        "--type",
        default="text_nd",
        help="input data format " "(text2d,text_nd(default),picture2d,pictures3d,npy)",
    )
    p.add_argument("-o", "--output", required=True, help="output complex file name")
    p.add_argument("-O", "--output-type", default="float", help="output data type (float, int, bool)")
    p.add_argument("-t", "--threshold", type=float, help="threshold for boolean output")
    add_argument_for_license(p)
    p.add_argument("input", nargs="*", help="input files")
    return p


def main(args=None):
    args = args or argument_parser().parse_args()
    pict = utils.load_nd_picture(args.input, args.type)
    if args.output_type in ["float", "int"]:
        pict = pict.astype(args.output_type)
    elif args.output_type == "bool":
        pict = pict > args.threshold

    np.save(args.output, pict)


if __name__ == "__main__":
    main()



import collections
import msgpack
import homcloud.optimal_one_cycle_ext as optimal_one_cycle_ext


def vertices(boundary_map, birth):
    return [index for (index, (dim, boundary)) in enumerate(boundary_map["map"][:birth]) if dim == 0]


def edges(boundary_map, birth):
    return [(index, boundary) for (index, (dim, boundary)) in enumerate(boundary_map["map"][:birth]) if dim == 1]


def adjacent_vertices(boundary_map, birth):
    adjacent_vertices = {vertex: [] for vertex in vertices(boundary_map, birth)}
    for e, (x, y) in edges(boundary_map, birth):
        adjacent_vertices[x].append((e, x, y))
        adjacent_vertices[y].append((e, y, x))

    return adjacent_vertices


def path(visited, start, end):
    v = end
    result = []

    while True:
        (edge, v) = visited[v]
        result.append(edge)
        if v is None:
            return result


def search(boundary_map, birth):
    adjacents = adjacent_vertices(boundary_map, birth)
    start, end = boundary_map["map"][birth][1]

    queue = collections.deque([(birth, None, start)])
    visited = {}

    while True:
        if len(queue) == 0:
            return None

        (e, v, other) = queue.popleft()
        if other in visited:
            continue
        visited[other] = (e, v)
        if other == end:
            return path(visited, start, end)

        queue.extend(adjacents[other])


def search_on_chunk_bytes(boundary_map_chunk_bytes, birth):
    return optimal_one_cycle_ext.search(boundary_map_chunk_bytes, birth)



import numpy as np


# TODO: Use KDTree or something
class SpatialSearcher(object):
    def __init__(self, pairs, births, deaths):
        self.pairs = pairs
        self.births = births
        self.deaths = deaths

    def nearest_pair(self, birth, death):
        distances = (self.births - birth) ** 2 + (self.deaths - death) ** 2
        return self.pairs[np.argmin(distances)]

    def in_rectangle(self, xmin, xmax, ymin, ymax):
        def is_pair_in_rectangle(birth, death):
            return (xmin <= birth <= xmax) and (ymin <= death <= ymax) and (birth != death)

        return [self.pairs[k] for k in range(len(self.pairs)) if is_pair_in_rectangle(self.births[k], self.deaths[k])]



from collections import deque
import numpy as np

from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import dijkstra


def search(adj_matrix, birth, weighted=False):
    if weighted:
        return search_for_weighted_graph(adj_matrix, birth)
    else:
        return Finder(adj_matrix, birth).search()


TERM = -1


def search_starting_edge(matrix, birth):
    xs, ys = np.nonzero(matrix == birth)
    if xs.size == 0:
        raise RuntimeError("Invalid birth time")
    if xs.size > 2:
        raise RuntimeError("Pairs with the same birth time")
    return sorted([xs[0], ys[0]])


def path(predecessors, start, end):
    result = [end]
    at = end
    while at != start:
        at = predecessors[at]
        result.append(at)
    return result


def search_for_weighted_graph(adjacent_matrix, birth):
    start, end = search_starting_edge(adjacent_matrix, birth)
    matrix = adjacent_matrix.copy()
    matrix[matrix >= birth] = 0.0
    graph = csr_matrix(matrix)
    dist_matrix, predecessors = dijkstra(graph, False, start, True)
    return path(predecessors, start, end)


def search_with_cocycle(adjacent_matrix, cocycle, birth, torelance, weighted):
    def edge(e):
        v1 = e[0]
        v2 = e[1]
        if v1 < v2:
            return (v1, v2)
        else:
            return (v2, v1)

    cocycle_edges = set(edge(e) for e in cocycle)
    matrix = adjacent_matrix.copy()
    matrix[matrix >= birth + torelance] = 0.0
    for e in cocycle_edges:
        matrix[e[0], e[1]] = 0.0
        matrix[e[1], e[0]] = 0.0
    graph = csr_matrix(matrix)

    # TODO: 最短距離計算を重複して計算しているので，効率化のためには保存しておくとよい
    predecessors = None
    min_loop_length = np.inf
    min_edge = None
    for e in cocycle_edges:
        if birth + torelance < adjacent_matrix[e[0], e[1]]:
            continue
        dist_matrix, another_predecessors = dijkstra(graph, False, e[0], True, not weighted)
        loop_length = dist_matrix[e[1]] + adjacent_matrix[e[0], e[1]]
        if min_loop_length > loop_length:
            min_loop_length = loop_length
            predecessors = another_predecessors
            min_edge = e
    return path(predecessors, min_edge[0], min_edge[1])


class Finder(object):
    def __init__(self, adjacent_matrix, birth):
        self.matrix = adjacent_matrix
        self.birth = birth
        self.visited = set()
        self.prev = dict()

    def search(self):
        start, end = search_starting_edge(self.matrix, self.birth)
        g = self.matrix < self.birth
        queue = deque([start])
        self.visited.add(start)
        self.prev[start] = TERM

        while queue:
            v = queue.popleft()
            if v == end:
                return self.path(end)
            for n in range(self.num_points):
                if not g[n, v]:
                    continue
                if n in self.visited:
                    continue

                self.visited.add(n)
                self.prev[n] = v
                queue.append(n)

        raise RuntimeError("No loop containing birth")

    def path(self, end):
        v = end
        path = []
        while True:
            path.append(v)
            v = self.prev[v]
            if v == TERM:
                return path

    @property
    def num_points(self):
        return self.matrix.shape[0]



import numpy as np

import homcloud.phtrees as phtrees
import homcloud.paraview_interface as pv_interface
import homcloud.plotly_3d as p3d
from homcloud.spatial_searcher import SpatialSearcher


class PHTrees(object):
    """
    This class represents PH trees computed from an alpha filtration.

    Please see `<https://arxiv.org/abs/1712.05103>`_ if you want to know
    the details of optimal volumes and volume optimal cycles.

    You can compute the PH trees by :meth:`PDList.from_alpha_filtration`
    with ``save_boundary_map=True`` and ``save_phtrees=True`` arguments.

    .. You can compute the PH trees by :meth:`PD.phtrees` from the diagram.

    Example:
        >>> import homcloud.interface as hc
        >>> pointcloud = hc.example_data("tetrahedron")
        >>> # Compute PDs and PHTrees
        >>> pdlist = hc.PDList.from_alpha_filtration(pointcloud, save_boundary_map=True, save_phtrees=True)
        >>> # Load phtrees
        >>> phtrees = pdlist.dth_diagram(2).load_phtrees()
        >>> # Query the node whose birth-death pair is nearest to (19, 21).
        >>> node = phtrees.pair_node_nearest_to(19, 21)
        >>> # Show birth time and death time
        >>> node.birth_time()
        19.600000000000005
        >>> node.death_time()
        21.069444444444443
        >>> node.boundary_points()
        [[0.0, 0.0, 0.0], [8.0, 0.0, 0.0], [5.0, 6.0, 0.0], [4.0, 2.0, 6.0]]
        >>> node.boundary()
        [[[0.0, 0.0, 0.0], [5.0, 6.0, 0.0], [4.0, 2.0, 6.0]],
         [[0.0, 0.0, 0.0], [8.0, 0.0, 0.0], [5.0, 6.0, 0.0]],
         [[8.0, 0.0, 0.0], [5.0, 6.0, 0.0], [4.0, 2.0, 6.0]],
         [[0.0, 0.0, 0.0], [8.0, 0.0, 0.0], [4.0, 2.0, 6.0]]]

    """

    def __init__(self, orig, spatial_searcher):
        self.orig = orig
        self.spatial_searcher = spatial_searcher

    @staticmethod
    def from_pdgm(pdgm):
        return PHTrees(
            phtrees.PHTrees.from_pdgm(pdgm, PHTrees.Node),
            SpatialSearcher(pdgm.death_indices, pdgm.births, pdgm.deaths),
        )

    @property
    def all_nodes(self):
        """
        Return all nodes.

        Returns:
            list[:class:`PHTrees.Node`]: The nodes.
        """
        return list(self.orig.nodes.values())

    @property
    def roots(self):
        return self.orig.roots

    def nodes_of(self, pairs):
        """
        Returns the nodes of trees corresponding to birth-death pairs
        in `pairs`.

        Args:
            pairs (list[Pair]): The list of pairs.

        Returns:
            list[:class:`PHTrees.Node`]: The nodes.
        """
        return [self._resolver().phtree.nodes[pair.death_index] for pair in pairs]

    def pair_node_nearest_to(self, x, y):
        """
        Return the node corresponding the pair which is nearest to
        (`x`, `y`).

        Args:
            x (float): The birth-axis coordinate.
            y (float): The death-axis coordinate.

        Returns:
            :class:`PHTrees.Node`: The nearest node.

        """
        return self.orig.nodes[self.spatial_searcher.nearest_pair(x, y)]

    def pair_nodes_in_rectangle(self, xmin, xmax, ymin, ymax):
        """
        Returns the list of nodes corresponding to the birth-death
        pairs in the given rectangle.

        Args:
           xmin (float): The minimum of the birth-axis of the rectangle.
           xmax (float): The maximum of the birth-axis of the rectangle.
           ymin (float): The minimum of the death-axis of the rectangle.
           ymax (float): The maximum of the death-axis of the rectangle.

        Returns:
           list[:class:`PHTrees.Node`]: The nodes in the rectangle.

        """
        return [
            self.orig.nodes[death_index] for death_index in self.spatial_searcher.in_rectangle(xmin, xmax, ymin, ymax)
        ]

    def to_paraview_node_from_nodes(self, nodes, gui_name=None):
        """
        Construct a :class:`homcloud.paraview_interface.PipelineNode`
        object to visulize optimal volumes of the nodes.

        Args:
            nodes (list[:class:`PHTrees.Node`]): The list of nodes to be visualized.
            gui_name (str | None): The name shown in Pipeline Browser
                in paraview's GUI.

        Returns:
            homcloud.paraview_interface.VTK: Paraview Pipeline node object.

        Notes:
            All nodes should be nodes of `self` PHTrees.
        """
        f = pv_interface.TempFile(".vtk")
        drawer = self.orig.coord_resolver.build_paraview_drawer(len(nodes), {})
        for i, node in enumerate(nodes):
            node.draw_volume(drawer, i)
        drawer.write(f)
        f.close()
        return pv_interface.VTK(f.name, gui_name, f).set_representation("Wireframe")

    to_pvnode_from_nodes = to_paraview_node_from_nodes

    class Volume(object):
        """
        The superclass of :class:`PHTrees.Node` and :class:`PHTrees.StableVolume`.

        Methods:
            birth_time()
                Returns:
                    float: The birth time of the corresponding birth-death pair.

            death_time()
                Returns:
                    float: The death time of the corresponding birth-death pair.

            lifetime()
                Returns:
                    float: The lifetime of the corresponding birth-death pair.

            simplices()
                Returns:
                    list[list[list[float]]], a.k.a list[Simplex]:
                        The simplices in the optimal volume.

            boundary()
                Returns:
                    list[list[float]], a.k.a. list[Point]:
                        Points in the volume optimal cycle.

            birth_simplex()
                Returns the birth simplex.

            death_simplex()
                Returns the death simplex.

            ancestors()
                Returns:
                    list[:class:`PHTrees.Node`]:
                        The ancestors of the tree node include itself.
        """

        def points(self):
            """
            Returns:
                list[list[float]], a.k.a list[Point]:
                    Points in the optimal volume.
            """
            return self.vertices()

        def volume(self):
            return self.volume_nodes

        def boundary_points(self):
            """
            Returns:
                list[list[float]]: All vertices in the boundary of the optimal/stable volume
            """
            return self.boundary_vertices()

        def points_symbols(self):
            """
            Returns:
                list[str]: All vertices in the optimal/stable volume
                in the form of the symbolic representation.
            """
            return self.vertices("symbols")

        def volume_simplices_symbols(self):
            """
            Returns:
                list[list[str]]: All simplices in optimal/stable volume
                in the form of the symbolic representation.
            """
            return self.simplices("symbols")

        def boundary_points_symbols(self):
            """
            Returns:
                list[str]: All vertices in the boundary of the optimal/stable volume
                in the form of the symbolic representation.
            """
            return self.boundary_vertices("symbols")

        def boundary_symbols(self):
            """
            Returns:
                list[list[str]]: All simplices in the optimal/stable volume
                in the form of the symbolic representation.
            """
            return self.boundary("symbols")

        def living(self):
            """
            Returns:
                bool: True if the birth time and death time of the node
                are different.
            """
            return self.birth_time() != self.death_time()

        def to_paraview_node(self, gui_name=None):
            """
            Construct a :class:`homcloud.paraview_interface.PipelineNode`
            object to visulize an optimal/stable volume of the node.

            gui_name (str | None): The name shown in Pipeline Browser
                in paraview's GUI.

            Returns:
                homcloud.paraview_interface.VTK: Paraview Pipeline node object.
            """
            f = pv_interface.TempFile(".vtk")
            drawer = self.trees.coord_resolver.build_paraview_drawer(1, {})
            self.draw_volume(drawer, 0)
            drawer.write(f)
            f.close()
            return pv_interface.VTK(f.name, gui_name, f).set_representation("Wireframe")

        to_pvnode = to_paraview_node

        def to_plotly3d_trace(self, color="green", width=1, name=""):
            """
            Constructs a plotly's trace object to visualize the optimal volume

            Args:
                color (str | None): The name of the color
                width (int): The width of the line
                name (str): The name of the object

            Returns:
                plotly.graph_objects.Scatter3d: Plotly's trace object
            """
            return p3d.Simplices(self.boundary(), color, width, name)

        to_plotly3d = to_plotly3d_trace

        def to_plotly3d_mesh(self, color="green", name=""):
            """
            Constructs a plotly's trace object to visualize the face of an optimal/stable volume

            Args:
                color (str | None): The name of the color
                name (str): The name of the object

            Returns:
                plotly.graph_objects.Mesh3d: Plotly's trace object
            """
            return p3d.SimplicesMesh(self.boundary(), color, name)

        def to_pyvista_mesh(self):
            """
            Constructs a PyVista's mesh object to visualize the face of an optimal/stable volume

            Returns:
                pyvista.PolyData: PyVista's mesh object
            """
            import homcloud.pyvistahelper as pvhelper

            return pvhelper.Triangles(self.boundary())

    class Node(Volume, phtrees.Node):
        """
        The class represents a tree node of :class:`PHTrees`.
        """

        def ancestors(self):
            ret = [self]
            while True:
                parent = self.trees.parent_of(ret[-1])
                if parent is None:
                    break
                ret.append(parent)

            return ret

        def living_descendants(self):
            """
            Returns:
                list[:class:`PHTrees.Node`]: All descendant nodes with positive lifetime
            """
            return [node for node in self.volume_nodes if node.living()]

        def stable_volume(self, epsilon):
            """
            Args:
                epsilon (float): Duration noise strength

            Returns:
                :class:`PHTrees.StableVolume`: The stable volume

            """
            return super().stable_volume(epsilon, PHTrees.StableVolume)

        def stable_volume_information(self):
            thresholds = np.array([child.birth_time() - self.birth_time() for child in self.children])
            volumes = np.array([child.volume_size() for child in self.children])
            return (thresholds, volumes)

        def __repr__(self):
            return "PHTrees.Node({}, {})".format(self.birth_time(), self.death_time())

    class StableVolume(Volume, phtrees.StableVolume):
        """
        The class represents a stable volume in :class:`PHTrees`.
        """

        pass



import homcloud.pict.optimal_one_cycle as pict_opt1cyc
import homcloud.paraview_interface as pv_interface
import homcloud.plotly_3d as p3d


class BitmapOptimal1Cycle(object):
    """The class represents an optimal (not volume-optimal) 1-cycle for bitmap.

    Computing volume-optimal cycle is very expensive for 3-D and
    higher dimensional cubical filtration. To fight against such
    a huge filtration, :meth:`Pair.optimal_1_cycle` is available.
    This method returns an instance of this class.
    """

    def __init__(self, orig):
        self.orig = orig

    def birth_time(self):
        """
        Returns:
            float: The birth time.
        """
        return self.orig.birth_time

    def death_time(self):
        """
        Returns:
            float: The death time.
        """
        return self.orig.death_time

    def birth_position(self):
        """
        Returns:
            tuple of float*N: The coordinate of birth position. (N: dimension)
        """
        return self.orig.path[0]

    def path(self):
        """
        Returns the path (loop) of the optimal 1-cycle.

        The first item and the last item is the same as :meth:`birth_position`.

        Returns:
            list of coord: The list of vertices of the loop ordered by the path
        """
        return self.orig.path

    def boundary_points(self):
        """
        Returns:
            list of coord: The list of vertices in the loop. Any vertex
                in the list is unique.
        """
        return self.orig.boundary_points()

    def to_paraview_node(self, gui_name=None):
        """
        Construct a :class:`homcloud.paraview_interface.PipelineNode` object
        to visulize an optimal 1-cycle.

        You can show the optimal 1-cycle by
        :meth:`homcloud.paraview_interface.show`. You can also
        adjust the visual by the methods of
        :class:`homcloud.paraview_interface.PipelineNode`.

        Args:
            gui_name (string or None): The name shown in Pipeline Browser
                in paraview's GUI.

        Returns:
            homcloud.paraview_interface.PipelineNode: A PipelineNode object.
        """
        return self.to_paraview_node_for_1cycles([self], gui_name)

    @staticmethod
    def to_paraview_node_for_1cycles(cycles, gui_name=None):
        """
        Construct a :class:`homcloud.paraview_interface.PipelineNode` object
        to visulize multiple optimal 1-cycles.

        Args:
            cycles (list of :class:`Optimal1CycleForBitmap`):
                The optimal 1-cycles to be visualized.
            gui_name (string or None): The name shown in Pipeline Browser
                in paraview's GUI.

        Returns:
            homcloud.paraview_interface.PipelineNode: A PipelineNode object.
        """

        drawer = pict_opt1cyc.prepare_drawer_for_paraview(len(cycles))
        for i, cycle in enumerate(cycles):
            cycle.orig.draw(drawer, i, str(i))
        f = pv_interface.TempFile(".vtk")
        drawer.write(f)
        f.close()
        return pv_interface.VTK(f.name, gui_name, f).set_representation("Wireframe")

    to_pvnode = to_paraview_node
    to_pvnode_for_1cycle = to_paraview_node_for_1cycles

    def to_plotly3d_trace(self, color=None, name=""):
        """
        Constructs a plotly's trace object to visualize the optimal 1-cycle

        Args:
            color (string or None): The name of the color
            name (string): The name of the object

        Returns:
            plotly.graph_objects.Mesh3d: Plotly's trace object
        """
        return p3d.Voxels(self.path(), color, name)

    to_plotly3d = to_plotly3d_trace

    def to_pyvista_mesh(self):
        """
        Constructs a PyVista's mesh object to visualize the optimal 1-cycle

        Returns:
            pyvista.PolyData: PyVista's mesh object
        """
        import homcloud.pyvistahelper as pvhelper

        return pvhelper.SparseVoxels(self.path())



import math

import homcloud.hera_bottleneck as hera_bottleneck
import homcloud.hera_wasserstein as hera_wasserstein


def bottleneck(pd1, pd2, delta=0.000001):
    """
    Compute the bottleneck distance between two diagrams.

    The parameter `delta` determines the acceptable relative error.
    If delta is zero, the return value is exact but slower.
    If delta is positive, the return value is not exact but faster.

    Notes:
        This function uses hera library <https://bitbucket.org/grey_narn/hera>.
        See the following paper for theoretical details:
        Michael Kerber, Dmitriy Morozov, and Arnur Nigmetov,
        "Geometry Helps to Compare Persistence Diagrams.",
        Journal of Experimental Algorithmics, vol. 22, 2017, pp. 1--20.
        (conference version: ALENEX 2016).

    Args:
        pd1 (PD): A persistence diagram
        pd2 (PD): Another persistence diagram
        delta (float): Acceptable relative error, must be zero or positive

    Returns:
        float: the bottleneck distance betweeen two diagrams
    """
    return hera_bottleneck.bottleneck_distance(pd1.births, pd1.deaths, pd2.births, pd2.deaths, delta)


def wasserstein(pd1, pd2, power=2, internal_p=math.inf, delta=0.000001):
    """
    Compute the Wasserstein distance between two diagrams.

    Example:
        >>> import homcloud.interface as hc
        >>> hc.distance.wasserstein(pd1, pd2, delta=0.01)
        Returns 2-Wasserstein distance with relative error <= 0.01

    Notes:
        This function uses hera library <https://bitbucket.org/grey_narn/hera>.
        See the following paper for theoretical details:
        Michael Kerber, Dmitriy Morozov, and Arnur Nigmetov,
        "Geometry Helps to Compare Persistence Diagrams.",
        Journal of Experimental Algorithmics, vol. 22, 2017, pp. 1--20.
        (conference version: ALENEX 2016).

    Args:
        pd1 (PD): A persistence diagram
        pd2 (PD): Another persistence diagram
        power (float): Wasserstein degree, must be larger than or equal to 1
        internal_p (float): The internal norm in Wasserstein distance,
            must be larger than or equal to 1 including infinity
        delta (float): Acceptable relative error, must be zero or positive

    Returns:
        float: the Wasserstein distance betweeen two diagrams
    """
    return hera_wasserstein.wasserstein_distance(
        pd1.births, pd1.deaths, pd2.births, pd2.deaths, power, internal_p, delta
    )



import math

import numpy as np
from sklearn import mixture


class MixedGaussians2(object):
    """Represents 2 mixed gaussians."""

    def __init__(self, gaussians):
        self.gaussians = gaussians
        self.boundary = self.find_boundary()

    def find_boundary(self):
        """Compute the value that which density is larger is changing."""
        (g0, g1) = self.gaussians
        for k in np.linspace(g0.mean, g1.mean, 256):
            if g0.density(k) < g1.density(k):
                return k
        return None

    @staticmethod
    def fit(pixels):
        """Fitting the values of pixels by 2-components GMM."""
        gmm = mixture.GMM(n_components=2)
        gmm.fit(pixels.reshape(pixels.size, 1))
        return MixedGaussians2(WeightedNormal.from_gmm(gmm))


class WeightedNormal(object):
    """This class represents weighted normal distribution.

    If weight is 1.0, the instance represents normal gaussian distribution.
    """

    def __init__(self, mean, var, weight):
        self.mean = mean
        self.var = var
        self.sd = math.sqrt(var)
        self.weight = weight

    def density(self, x):
        """Returns the density at x (in float)"""
        from scipy import stats

        return self.weight * stats.norm.pdf(x, self.mean, self.sd)

    @staticmethod
    def from_gmm(gmm):
        """Create a list of WeightedNormal objecsts from sklearn.mixture.GMM."""
        means = gmm.means_.flatten()
        variances = gmm.covars_.flatten()
        weights = gmm.weights_
        dists = [WeightedNormal(m, v, w) for (m, v, w) in zip(means, variances, weights)]
        dists.sort(key=lambda d: d.mean)
        return dists



import numpy as np
import numbers


def ary2v3(ary):
    return {"x": ary[0], "y": ary[1], "z": ary[2]}


def add_edges(view, edges, color, radius, alpha=1.0):
    assert isinstance(color, str)
    assert isinstance(radius, numbers.Real)
    assert isinstance(alpha, numbers.Real)

    for edge in edges:
        view.addCylinder(
            {"start": ary2v3(edge[0]), "end": ary2v3(edge[1]), "color": color, "radius": radius, "alpha": alpha}
        )


def add_surface(view, triangles, color, alpha=1.0):
    assert isinstance(color, str)
    assert isinstance(alpha, numbers.Real)

    for cell in triangles:
        normal = ary2v3(np.cross(np.array(cell[1]) - cell[0], np.array(cell[2]) - cell[0]))
        view.addCustom(
            {
                "vertexArr": [ary2v3(cell[0]), ary2v3(cell[1]), ary2v3(cell[2])],
                "normalArr": [normal, normal, normal],
                "faceArr": [0, 1, 2],
                "color": color,
                "alpha": alpha,
            }
        )



import itertools
from tempfile import TemporaryDirectory
import os
import math
from collections import defaultdict

import numpy as np
from forwardable import forwardable

import homcloud.utils as utils


class ParaViewColors(object):
    def __init__(self, n_colors):
        self.n_colors = n_colors
        self.color_scalars = np.linspace(0, 1.0, n_colors + 3)

    @property
    def various_colors(self):
        return self.color_scalars[:-3]

    def birth_color(self):
        return self.color_scalars[-2]

    def death_color(self):
        return self.color_scalars[-1]

    def output_lookup_table(self, f):
        f.write("LOOKUP_TABLE color_table {}\n".format(self.n_colors + 3))
        for x in np.linspace(0.0, 1.0, self.n_colors + 1):
            self.output_various_colors(f, x)
        self.output_birthdeath_colors(f)

    def output_various_colors(self, f, relative_value):
        f.write("{} {} {} 1.0\n".format(*self.color_spec(relative_value)))

    @staticmethod
    def color_spec(x):
        y, k = math.modf(x * 6)
        if k == 0 or k == 6:
            return (1, y, 0)
        if k == 1:
            return (1 - y, 1, 0)
        if k == 2:
            return (0, 1, y)
        if k == 3:
            return (0, 1 - y, 1)
        if k == 4:
            return (y, 0, 1)
        if k == 5:
            return (1, 0, 1 - y)
        return (0.5, 0.5, 0.5)

    @staticmethod
    def output_birthdeath_colors(f):
        f.write("0.8 0.8 0.8 1.0\n")
        f.write("1.0 1.0 1.0 1.0\n")


# TODO: this class requires unit tests
@forwardable()
class ParaViewDrawer(object):
    """
    This class represents the generator of a VTK file.

    Args:
        n_colors (int): The number of colors
        column_spec (dict[str, float or None]): The pairs of the names of
            the columns and default values.
    """

    def __init__(self, n_colors, column_spec):
        self.pvcolors = ParaViewColors(n_colors)
        self.lines = []
        self.colors = []
        self.columns = {name: list() for name in column_spec}
        self.default_values = {
            name: default_val for (name, default_val) in column_spec.items() if default_val is not None
        }
        self.vertices = []

    def_delegators("pvcolors", "various_colors,birth_color,death_color")

    def draw_line(self, p, q, color, **threshold_values):
        self.lines.append((p, q))
        self.append_attributes(color, threshold_values)

    def append_attributes(self, color, threshold_values):
        self.colors.append(color)
        for key in self.columns:
            self.columns[key].append(self.get_threshold_value(threshold_values, key))

    def get_threshold_value(self, values, key):
        if key in values:
            return values[key]
        else:
            return self.default_values[key]

    def invoke(self, wait=True):
        with TemporaryDirectory() as tmpdir:
            path = os.path.join(tmpdir, "tmp.vtk")
            self.output(path)
            utils.invoke_paraview(path, wait=wait)

    def output(self, path):
        with open(path, "w") as f:
            self.write(f)

    def write(self, f):
        self.output_header(f)
        self.output_polygon_data(f)
        self.output_line_colors(f)
        self.pvcolors.output_lookup_table(f)
        self.output_columns(f)

    @staticmethod
    def output_header(f):
        f.write("# vtk DataFile Version 2.0\n")
        f.write("volume optimal cycles\n")
        f.write("ASCII\n")

    def output_polygon_data(self, f):
        f.write("DATASET POLYDATA\n")
        self.output_points(f)
        self.output_vertices(f)
        self.output_lines(f)
        self.output_polygons(f)

    def output_vertices(self, f):
        if not self.vertices:
            return
        f.write("VERTICES {} {}\n".format(self.num_vertices(), self.num_vertices() * 2))
        for v in self.vertices:
            f.write("1 {}\n".format(v))
        f.write("\n")

    def output_lines(self, f):
        pass

    def output_polygons(self, f):
        pass

    def output_line_colors(self, f):
        f.write("CELL_DATA {}\n".format(len(self.colors)))
        f.write("SCALARS colors float 1\n")
        f.write("LOOKUP_TABLE color_table\n")
        for n in self.colors:
            f.write("{}\n".format(n))

    def output_columns(self, f):
        for key, values in self.columns.items():
            f.write("SCALARS {} float 1\n".format(key))
            f.write("LOOKUP_TABLE default\n")
            for value in values:
                f.write("{}\n".format(value))

    def num_lines(self):
        return len(self.lines)

    def num_vertices(self):
        return len(self.vertices)

    @staticmethod
    def reformat_point(point):
        if len(point) == 3:
            return point
        if len(point) == 2:
            return (point[0], point[1], 0)


class ParaViewSimplexDrawer(ParaViewDrawer):
    def __init__(self, n_colors, points, column_names):
        super().__init__(n_colors, column_names)
        self.points = [self.reformat_point(p) for p in points]

    def draw_simplex(self, simplex, color, **threshold_values):
        for edge in itertools.combinations(simplex, 2):
            self.draw_line(edge[0], edge[1], color, **threshold_values)

    def draw_vertex(self, vertex, color, **threshold_values):
        self.vertices.append(vertex)
        self.append_attributes(color, threshold_values)

    def draw_all_vertices(self, color, **threshold_values):
        for i in range(len(self.points)):
            self.draw_vertex(i, color, **threshold_values)

    def output_points(self, f):
        f.write("POINTS {} double\n".format(self.num_points()))
        for point in self.points:
            f.write("{} {} {}\n".format(point[0], point[1], point[2]))
        f.write("\n")

    def output_lines(self, f):
        f.write("LINES {} {}\n".format(self.num_lines(), self.num_lines() * 3))
        for line in self.lines:
            f.write("2 {} {}\n".format(line[0], line[1]))
        f.write("\n")

    def num_points(self):
        return len(self.points)


class ParaViewCubeDrawer(ParaViewDrawer):
    def __init__(self, n_colors, dims, column_names):
        super().__init__(n_colors, column_names)
        self.dims = dims
        assert len(dims) in [2, 3]

    @staticmethod
    def dvs(non_deg):
        n = len(non_deg)
        dvs = []

        def iter(k, dv):
            if k == n:
                dvs.append(dv.copy())
            else:
                if non_deg[k]:
                    for b in [0, 1]:
                        dv[k] = b
                        iter(k + 1, dv)
                else:
                    iter(k + 1, dv)

        iter(0, np.zeros(n, dtype=int))
        return dvs

    @staticmethod
    def dls(non_deg):
        n = len(non_deg)
        dls = []
        for k in range(n):
            if non_deg[k]:
                dl = np.zeros(n, dtype=int)
                dl[k] = 1
                dls.append(dl)
        return dls

    def draw_cube(self, coord, non_deg, color, **threshold_values):
        for dv in self.dvs(non_deg):
            for dl in self.dls(non_deg):
                if np.max(dv + dl) < 2:
                    self.draw_line(coord + dv, coord + dv + dl, color, **threshold_values)

    def ndim(self):
        return len(self.dims)

    def coord2index(self, coord):
        index = 0
        for k in range(self.ndim()):
            index = index * self.dims[k] + coord[k]
        return index

    def index2coord(self, index):
        coord = [0] * self.ndim()
        for k in reversed(range(self.ndim())):
            index, coord[k] = divmod(index, self.dims[k])
        return coord

    def output_points(self, f):
        f.write("POINTS {} double\n".format(self.num_points()))
        for k in range(self.num_points()):
            coord = self.index2coord(k)
            f.write("{} {} {}\n".format(*self.reformat_point(coord)))
        f.write("\n")

    def num_points(self):
        return int(np.prod(self.dims))

    def output_lines(self, f):
        f.write("LINES {} {}\n".format(self.num_lines(), self.num_lines() * 3))
        for line in self.lines:
            f.write("2 {} {}\n".format(self.coord2index(line[0]), self.coord2index(line[1])))
        f.write("\n")


class ParaViewSparseCubeDrawer(ParaViewDrawer):
    def __init__(self, n_colors, ndim, column_names):
        super().__init__(n_colors, column_names)
        self.ndim = ndim
        self.points = []
        assert ndim in [2, 3]

    def draw_cube(self, coord, non_deg, color, **threshold_values):
        indices, dvs = self.prepare_points(coord, non_deg)
        for dv in dvs:
            for dl in ParaViewCubeDrawer.dls(non_deg):
                if np.max(dv + dl) < 2:
                    self.draw_line(indices[tuple(dv)], indices[tuple(dv + dl)], color, **threshold_values)

    def prepare_points(self, coord, non_deg):
        dvs = ParaViewCubeDrawer.dvs(non_deg)
        indices = np.zeros([2] * self.ndim, dtype=int)
        for dv in dvs:
            indices[tuple(dv)] = len(self.points)
            self.points.append(np.flip(dv + coord))
        return indices, dvs

    def output_points(self, f):
        f.write("POINTS {} double\n".format(self.num_points()))
        for point in self.points:
            f.write("{} {} {}\n".format(point[0], point[1], point[2]))
        f.write("\n")

    def num_points(self):
        return len(self.points)

    def output_lines(self, f):
        f.write("LINES {} {}\n".format(self.num_lines(), self.num_lines() * 3))
        for line in self.lines:
            f.write("2 {} {}\n".format(line[0], line[1]))
        f.write("\n")


class ParaViewSparseBitmapDrawer(ParaViewDrawer):
    def __init__(self, n_colors, column_names):
        super().__init__(n_colors, column_names)
        self.vertices = []
        self.n_voxels = 0

    def draw_voxel(self, coord, color, **threshold_values):
        self.n_voxels += 1
        for d1 in (-0.4, 0.4):
            for d2 in (-0.4, 0.4):
                for d3 in (-0.4, 0.4):
                    r = (coord[0] + d1, coord[1] + d2, coord[2] + d3)
                    self.vertices.append(r)
        self.append_attributes(color, threshold_values)

    def output(self, path):
        with open(path, "w") as f:
            self.output_header(f)
            self.output_voxel_data(f)
            self.output_line_colors(f)
            self.pvcolors.output_lookup_table(f)
            self.output_columns(f)

    def output_voxel_data(self, f):
        f.write("DATASET UNSTRUCTURED_GRID\n")
        f.write("POINTS {} double\n".format(len(self.vertices)))
        for r in self.vertices:
            f.write(" ".join(map(str, r)))
            f.write("\n")
        f.write("CELLS {} {}\n".format(self.n_voxels, 9 * self.n_voxels))
        for n in range(self.n_voxels):
            f.write("8 {}\n".format(" ".join(map(str, range(n * 8, (n + 1) * 8)))))
        f.write("CELL_TYPES {}\n".format(self.n_voxels))
        for _ in range(self.n_voxels):
            f.write("11\n")


class ParaViewLinesDrawer(ParaViewDrawer):
    def __init__(self, n_colors, column_names):
        super().__init__(n_colors, column_names)
        self.points = []
        self.point2index = dict()
        self.lines = []

    def index_of(self, p):
        if p in self.point2index:
            return self.point2index[p]
        else:
            index = len(self.points)
            self.point2index[p] = index
            self.points.append(p)
            return index

    def draw_line(self, p1, p2, color, **threshold_values):
        index1 = self.index_of(p1)
        index2 = self.index_of(p2)
        self.lines.append((index1, index2))
        self.append_attributes(color, threshold_values)

    def output_points(self, f):
        f.write("POINTS {} double\n".format(self.num_points()))
        for point in self.points:
            point = self.reformat_point(point)
            f.write("{} {} {}\n".format(point[0], point[1], point[2]))
        f.write("\n")

    def num_points(self):
        return len(self.points)

    def output_lines(self, f):
        f.write("LINES {} {}\n".format(self.num_lines(), self.num_lines() * 3))
        for line in self.lines:
            f.write("2 {} {}\n".format(line[0], line[1]))
        f.write("\n")

    def draw_loop(self, path, color, **threshold_values):
        for k in range(len(path) - 1):
            self.draw_line(path[k], path[k + 1], color, **threshold_values)


class ParaViewPolyLineDrawer(ParaViewDrawer):
    def __init__(self, n_colors, points, column_names):
        super().__init__(n_colors, column_names)
        self.points = [self.reformat_point(p) for p in points]
        self.append_attributes(0, {})

    def output_points(self, f):
        f.write("POINTS {} double\n".format(self.num_points()))
        for point in self.points:
            f.write("{} {} {}\n".format(point[0], point[1], point[2]))
        f.write("\n")

    def output_lines(self, f):
        f.write("LINES 1 {}\n".format(self.num_points() + 1))
        f.write("{} ".format(self.num_points()))
        for k in range(self.num_points()):
            f.write(" {}".format(k))
        f.write("\n\n")

    def num_points(self):
        return len(self.points)


class ParaView3DSurfaceDrawer(ParaViewDrawer):
    def __init__(self, n_colors, surface, column_names):
        super().__init__(n_colors, column_names)
        self.num_triagnles = len(surface)
        self.surface = surface
        for _ in range(self.num_triagnles):
            self.append_attributes(0, {})

    def output_points(self, f):
        points = np.array(self.surface).reshape(3 * self.num_triagnles, 3)
        f.write("POINTS {} double\n".format(self.num_triagnles * 3))
        for point in points:
            f.write("{} {} {}\n".format(point[0], point[1], point[2]))
        f.write("\n")

    def output_polygons(self, f):
        f.write("POLYGONS {} {}\n".format(self.num_triagnles, self.num_triagnles * 4))
        for k in range(self.num_triagnles):
            f.write("3 {} {} {}\n".format(3 * k, 3 * k + 1, 3 * k + 2))


def boundary_of_connected_simplices(simplices):
    def boundaries(key):
        if len(key) == 1:
            return []
        return [key.difference([el]) for el in key]

    count = defaultdict(int)
    for simplex in simplices:
        for boundary in boundaries(simplex):
            count[boundary] += 1

    return [simplex for simplex, num in count.items() if num == 1]



import tempfile
import os
import unittest.mock
import shutil
import sys
import re

import numpy as np
from scipy.spatial import distance_matrix

import homcloud.interface as hc
import homcloud.paraview_interface as pv
import homcloud.utils as utils
import homcloud.pict.pict3d_vtk as pict3d_vtk
from homcloud.version import __version__

import matplotlib

matplotlib.rcParams["backend"] = "Agg"

import matplotlib.pyplot as plt  # NOQA


def self_check(args):
    print_version()
    self_check_alpha_shape_3()
    self_check_alpha_shape_3_with_weight()
    self_check_alpha_shape_3_periodic_with_weight()
    self_check_alpha_shape_2()
    self_check_grayscale_2D_bitmap()
    self_check_binary_2D_bitmap()
    self_check_binary_periodic_2D_bitmap()
    self_check_rips()
    if "--dipha" in args:
        self_check_rips_maxvalue()
    self_check_plot_pd()
    self_check_optimal_volume()
    if "--paraview" in args:
        self_check_paraview()
    if "--plotly" in args:
        self_check_plotly()
    if "--pyvista" in args:
        self_check_pyvista()


def print_version():
    print("HomCloud version: {}".format(__version__))
    print("Python version: {}".format(re.sub(r"[\r\n]", " ", sys.version)))


def self_check_alpha_shape_3():
    print("Alpha Shape 3 ... ", end="")
    pc = np.random.uniform(0, 1.0, (100, 3))
    hc.PDList.from_alpha_filtration(pc)
    print("ok")


def self_check_alpha_shape_3_with_weight():
    print("Alpha Shape 3 with weights ... ", end="")
    pc = np.random.uniform(0, 1.0, (100, 3))
    weight = np.random.uniform(0.001, 0.05, 100)
    hc.PDList.from_alpha_filtration(pc, weight=weight)
    print("ok")


def self_check_alpha_shape_3_periodic_with_weight():
    print("Periodic Alpha Shape 3 with weights ... ", end="")
    pc = np.random.uniform(0, 1, (200, 3))
    weight = np.random.uniform(0.001, 0.05, 200)
    hc.PDList.from_alpha_filtration(pc, weight=weight, periodicity=[(0, 1), (0, 1), (0, 1)])
    print("ok")


def self_check_alpha_shape_2():
    print("Alpha Shape 2 ... ", end="")
    pc = np.random.uniform(0, 1.0, (100, 2))
    hc.PDList.from_alpha_filtration(pc)
    print("ok")


def self_check_grayscale_2D_bitmap():
    print("Grayscale 2D bitmap ... ", end="")
    bitmap = np.random.uniform(0, 1.0, (128, 128))
    hc.PDList.from_bitmap_levelset(bitmap)
    print("ok")


def self_check_binary_2D_bitmap():
    print("Binary 2D bitmap ... ", end="")
    bitmap = np.random.uniform(0, 1.0, (128, 128))
    hc.PDList.from_bitmap_levelset(hc.distance_transform(bitmap < 0.2, True, "manhattan"))
    print("ok")


def self_check_binary_periodic_2D_bitmap():
    print("Binary 2D periodic bitmap ... ", end="")
    bitmap = np.random.uniform(0, 1.0, (128, 128))
    hc.PDList.from_bitmap_levelset(hc.distance_transform(bitmap < 0.2, True, "manhattan", (True, True)))
    print("ok")


def self_check_rips():
    print("Rips filtration ... ", end="")
    pc = np.random.uniform(0, 1.0, (64, 128))
    dmatrix = distance_matrix(pc, pc)
    hc.PDList.from_rips_filtration(dmatrix, maxdim=2)
    print("ok")


def self_check_rips_maxvalue():
    print("Rips filtration with maxvalue ... ", end="")
    pc = np.random.uniform(0, 1.0, (64, 128))
    dmatrix = distance_matrix(pc, pc)
    hc.PDList.from_rips_filtration(dmatrix, maxdim=2, maxvalue=0.01)
    print("ok")


def self_check_plot_pd():
    print("Plotting PD ... ", end="")
    pc = np.random.uniform(0, 1.0, (100, 3))
    pdlist = hc.PDList.from_alpha_filtration(pc)
    with tempfile.TemporaryDirectory() as tmpdir:
        for d in [0, 1, 2]:
            plt.clf()
            pdlist.dth_diagram(d).histogram().plot(colorbar={"type": "log"})
            plt.savefig(os.path.join(tmpdir, "PD.png"))
    plt.clf()
    print("ok")


def self_check_optimal_volume():
    print("Optimal Volume ... ", end="")
    pc = np.array([[0.0, 0.0, 0.0], [8.0, 0.0, 0.0], [5.0, 6.0, 0.0], [4.0, 2.0, 6.0]])
    pdlist = hc.PDList.from_alpha_filtration(pc, save_boundary_map=True)
    pdlist[1].nearest_pair_to(16, 19).optimal_volume()
    pdlist[2].nearest_pair_to(19.6, 21).optimal_volume()
    print("ok")


def self_check_paraview():
    print("Paraview path: {0} ({1})".format(utils.paraview_programname(), shutil.which(utils.paraview_programname())))
    print("Paraview fake invoke ... ", end="")
    with unittest.mock.patch("subprocess.check_call"):
        with unittest.mock.patch("subprocess.Popen"):
            utils.invoke_paraview("foo.vtk", wait=False)
            utils.invoke_paraview("foo.vtk", wait=True)
    print("ok")

    print('Paraview real invoke (VTK voxel) => Click "Apply" button and close opened window ... ', end="")
    sys.stdout.flush()

    with tempfile.TemporaryDirectory() as tmpdir:
        path = os.path.join(tmpdir, "voxel.vtk")
        with open(path, "w") as f:
            pict3d_vtk.write_vtk_file(f, "Voxel data", np.random.uniform(size=(10, 10, 10)))
        utils.invoke_paraview(path, wait=True)
    print("ok")

    print("Paraview real invoke (python pointcloud) => Close opened window ... ", end="")
    sys.stdout.flush()
    pv.show(
        [
            pv.PointCloud.from_array(np.random.uniform(size=(100, 3))),
            pv.PointCloud.from_array(np.random.uniform(size=(100, 3))).set_color((1, 0, 0)),
        ],
        wait=True,
    )
    print("ok")


def self_check_plotly():
    import plotly.graph_objects as go
    import homcloud.plotly_3d as p3d

    print("Plotly 3D drawing (3D objects are displayed in your browser, check it and close the tab) ...", end="")
    vertices = [[0, 0, 0], [1, 0, 2], [-0.7, 1.4, 1.9], [-0.7, -1.4, 2.1]]
    tetrahedron = [
        [vertices[0], vertices[1], vertices[2]],
        [vertices[0], vertices[1], vertices[3]],
        [vertices[0], vertices[2], vertices[3]],
        [vertices[1], vertices[2], vertices[3]],
    ]
    loop = [[np.cos(2 * np.pi / 6 * n), np.sin(2 * np.pi / 6 * n), 0] for n in range(6)]
    bitmap3d = np.array([[[0, 0, 0], [0, 1, 0], [0, 1, 0]], [[1, 1, 1], [1, 1, 1], [1, 1, 0]]])
    fig = go.Figure(
        data=[
            p3d.Simplices(tetrahedron, color="black", width=2, name="tetrahedron"),
            p3d.SimplicesMesh(tetrahedron, color="blue"),
            p3d.Loop(loop, color="green", width=4, name="loop"),
            p3d.Bitmap3d(bitmap3d, color="green"),
        ],
        layout=dict(scene=p3d.SimpleScene()),
    )
    fig.update_traces(opacity=0.5, selector=dict(type="mesh3d"))
    fig.show(renderer="browser")

    print(" ok")


def self_check_pyvista():
    import pyvista as pv
    import homcloud.pyvistahelper as pvhelper

    print("PyVista 3D drawing (close the pop-up window)...", end="")
    pl = pv.Plotter(notebook=False)
    pl.add_mesh(pv.PointSet(np.random.uniform(-10, 10, (100, 3))))
    pl.add_mesh(pv.Sphere(2, (4.5, 4.5, 4.5)), opacity=0.5, color="blue")
    pl.add_mesh(pvhelper.Bitmap3D(np.random.uniform(0, 1, (5, 5, 5))), clim=(0, 1))
    pl.show()

    print(" ok")


if __name__ == "__main__":
    self_check(sys.argv[1:])



from functools import partial
import argparse
import re
from bisect import bisect
from pathlib import Path
import json

import numpy as np
import pulp
from forwardable import forwardable

from homcloud.argparse_common import parse_range, parse_bool
from homcloud.license import add_argument_for_license
from homcloud.version import __version__
from homcloud.pdgm import PDGM
from homcloud.spatial_searcher import SpatialSearcher
from homcloud.numpyencoder import NumpyEncoder
from homcloud.visualize_3d import ParaViewSimplexDrawer, ParaViewCubeDrawer

try:
    from pulp.solvers import PulpSolverError
except ImportError:
    from pulp.apis.core import PulpSolverError


def find_lp_solver(name, options):
    if name is None:
        return default_lp_solver()
    if name == "coin" or name == "coin-cmd":
        return pulp.COIN_CMD(options=options)
    if name == "cplex":
        return pulp.CPLEX(options=options)
    if name == "gurobi":
        return pulp.GUROBI(options=options)
    if name == "glpk":
        return pulp.GLPK(options=options)
    if name == "coinmp-dll":
        return pulp.COINMP_DLL(options=options)
    if name == "cplex-dll":
        return pulp.CPLEX_DLL(options=options)
    raise RuntimeError("Unknown solver {}".format(name))


def default_lp_solver():
    lp_solver = pulp.LpSolverDefault.copy()
    lp_solver.msg = 0
    return lp_solver


def sign_function(boundary_map):
    def sign_simplicial(i, kth):
        return -1 if kth % 2 else 1

    def sign_abstract(i, kth):
        return boundary_map["map"][i][2][kth]

    def sign_cubical(i, kth):
        return -1 if (kth % 4) in (1, 2) else 1

    if boundary_map["type"] == "simplicial":
        return sign_simplicial
    if boundary_map["type"] == "abstract":
        return sign_abstract
    if boundary_map["type"] == "cubical":
        return sign_cubical

    raise RuntimeError("Unknown maptype: {}".format(boundary_map["type"]))


class Optimizer(object):
    def __init__(self, birth, death, boundary_map, lp_solver):
        self.birth = birth
        self.death = death
        self.boundary_map = boundary_map
        self.lp_solver = lp_solver

    def build_partial_boundary_matrix(self, is_active_cell):
        map = self.boundary_map["map"]
        sign = sign_function(self.boundary_map)

        def dim(i):
            return map[i][0]

        degree = dim(self.death) - 1
        partial_map = dict()
        lpvars = []

        for i in range(self.birth, self.death + 1):
            if dim(i) == degree:
                partial_map[i] = list()
            elif dim(i) == degree + 1 and is_active_cell(i):
                lpvars.append(i)
                for kth, j in enumerate(map[i][1]):
                    if j in partial_map:
                        partial_map[j].append((sign(i, kth), i))

        return lpvars, partial_map

    def build_lp_problem(self, lpvar_indices, partial_map):
        prob = pulp.LpProblem("OptimalVolume", pulp.LpMinimize)
        xs = pulp.LpVariable.dicts("x", (lpvar_indices,), -1, 1, "Continuous")
        ys = pulp.LpVariable.dicts("y", (lpvar_indices,), 0, 1, "Continuous")

        prob.setObjective(pulp.lpSum(ys.values()))

        for i in lpvar_indices:
            prob.addConstraint(xs[i] - ys[i] <= 0.0)
            prob.addConstraint(xs[i] + ys[i] >= 0.0)

        for low, constraint in partial_map.items():
            if not constraint:
                continue
            if low == self.birth:
                continue
            prob.addConstraint(pulp.lpSum(s * xs[i] for (s, i) in constraint) == 0)

        prob.addConstraint(xs[self.death] == 1)

        return prob, ys

    def find(self, is_active_cell=lambda _: True):
        lpvars, partial_map = self.build_partial_boundary_matrix(is_active_cell)
        prob, ys = self.build_lp_problem(lpvars, partial_map)
        try:
            status = prob.solve(self.lp_solver)
        except PulpSolverError as err:
            # workaround for cplex and pulp
            if re.search("infeasible", err.args[0]):
                return Optimizer.Infeasible(pulp.LpStatusInfeasible)
            else:
                raise err

        if status == pulp.LpStatusOptimal:
            return Success(self.optimal_volume_cell_indices(ys))
        else:
            return Failure(status)

    @staticmethod
    def optimal_volume_cell_indices(ys):
        return [index for (index, var) in ys.items() if var.varValue >= 0.00001]


class Success(object):
    def __init__(self, cell_indices):
        self.cell_indices = cell_indices
        self.pair = None
        self.subvolume = None

    success = True
    infeasible = False

    def children(self, death, death_to_birth):
        def valid(c):
            return c != death and c in death_to_birth

        return [(death_to_birth[c], c) for c in self.cell_indices if valid(c)]

    def __repr__(self):
        return "<optvol.Success: {}>".format(self.cell_indices)

    def draw(self, drawer, coord_resolver, volume_color, boundary_color, index, issubvolume=0):
        coord_resolver.draw_cells(
            drawer, self.cell_indices, volume_color, index=index, issimplex=1, issubvolume=issubvolume
        )
        coord_resolver.draw_boundary_cells(
            drawer, self.cell_indices, boundary_color, index=index, isboundary=1, issubvolume=issubvolume
        )
        if self.subvolume:
            self.subvolume.draw(drawer, coord_resolver, drawer.various_colors[2], drawer.various_colors[3], index, 1)


class Failure(object):
    def __init__(self, status):
        self.status = status
        self.pair = None

    success = False

    @property
    def infeasible(self):
        # NOTE: GLPK solver returns pulp.LpStatusUndefined when a problem is infeasible.
        # Therefore pulp.LpStatusUndefined is treated as an infeasible failure.
        # This is an workaround for GLPK.
        return self.status in [pulp.LpStatusInfeasible, pulp.LpStatusUndefined]

    @property
    def message(self):
        msg = pulp.LpStatus[self.status]
        if self.pair:
            return "{} at ({}, {})".format(msg, *self.pair)
        else:
            return msg

    def __repr__(self):
        return "<optvol.Failure: {}>".format(self.message)

    def draw(self, drawer, coord_resolver, volume_color, boundary_color, index):
        pass


class RetryOptimizer(object):
    def __init__(self, optimizer, is_active_cell, times):
        self.optimizer = optimizer
        self.is_active_cell = is_active_cell
        self.times = times

    def find(self):
        for n in range(self.times):
            result = self.optimizer.find(partial(self.is_active_cell, n))
            if result.infeasible:
                pass
            else:
                return result

        return result  # This line always returns infeasible result


def is_active_cell(coord_resolver, center, threshold, n, cell_index):
    if threshold is None:
        return True
    centroid = np.array(coord_resolver.centroid(cell_index))
    return np.linalg.norm(center - centroid) < threshold * (2**n)


class OptimizerBuilder(object):
    def __init__(self, degree, boundary_map, lp_solver):
        self.degree = degree
        self.boundary_map = boundary_map
        self.lp_solver = lp_solver

    def build(self, birth, death):
        return Optimizer(birth, death, self.boundary_map, self.lp_solver)

    def to_query_dict(self):
        return {"degree": self.degree, "solver-name": self.solver_name, "solver-options": self.solver_options}

    @property
    def solver_name(self):
        return self.lp_solver.__class__.__name__

    @property
    def solver_options(self):
        if hasattr(self.lp_solver, "getOptions"):
            return self.lp_solver.getOptions()
        else:
            return None

    @staticmethod
    def from_alpha_pdgm(pdgm, cutoff_radius, retry, lp_solver):
        return RetryOptimizerBuilder(
            pdgm.degree, pdgm.boundary_map_chunk, lp_solver, pdgm.alpha_coord_resolver, cutoff_radius, retry
        )

    @staticmethod
    def from_cubical_pdgm(pdgm, cutoff_radius, retry, lp_solver):
        return RetryOptimizerBuilder(
            pdgm.degree, pdgm.boundary_map_chunk, lp_solver, pdgm.cubical_geometry_resolver, cutoff_radius, retry
        )

    def builder_without_retry(self):
        return self


class RetryOptimizerBuilder(OptimizerBuilder):
    def __init__(self, degree, boundary_map, lp_solver, coord_resolver, cutoff, retry):
        super().__init__(degree, boundary_map, lp_solver)
        self.coord_resolver = coord_resolver
        self.cutoff = cutoff
        self.retry = retry

    def build(self, birth, death):
        return RetryOptimizer(
            super().build(birth, death),
            partial(is_active_cell, self.coord_resolver, self.coord_resolver.centroid(death), self.cutoff),
            self.retry,
        )

    def to_query_dict(self):
        return dict(
            super().to_query_dict(),
            **{
                "cutoff-radius": self.cutoff,
                "num-retry": self.retry,
            }
        )

    def builder_without_retry(self):
        return OptimizerBuilder(self.degree, self.boundary_map, self.lp_solver)


def bisect_tightened(b, d, index_to_level, epsilon):
    k = bisect(index_to_level, index_to_level[b] + epsilon, lo=b, hi=d + 1)
    assert k <= d, "epsilon must be smaller than lifetime"
    # 次の行で k - 1 を使っているのは，Optimizer.build_lp_problemで
    # if low == self.birth:
    #      continue
    # となっている部分と整合性を取るため
    return k - 1


class TightenedVolumeFinder(object):
    def __init__(self, optimizer_builder, index_to_level, epsilon):
        self.optimizer_builder = optimizer_builder
        self.index_to_level = index_to_level
        self.epsilon = epsilon

    def find(self, birth, death):
        k = bisect_tightened(birth, death, self.index_to_level, self.epsilon)
        return self.optimizer_builder.build(k, death).find()

    def to_query_dict(self):
        return dict(self.optimizer_builder.to_query_dict(), **{"query-target": "tightened-volume"})


class OptimalVolumeFinder(object):
    def __init__(self, optimizer_builder):
        self.optimizer_builder = optimizer_builder

    def find(self, birth, death):
        return self.optimizer_builder.build(birth, death).find()

    def to_query_dict(self):
        return dict(self.optimizer_builder.to_query_dict(), **{"query-target": "optimal-volume"})


class TightenedSubVolumeFinder(object):
    def __init__(self, optimizer_builder, index_to_level, epsilon):
        self.optimizer_builder = optimizer_builder
        self.index_to_level = index_to_level
        self.epsilon = epsilon

    def find(self, birth, death, optimal_volume_cells):
        k = bisect_tightened(birth, death, self.index_to_level, self.epsilon)
        return self.optimizer_builder.build(k, death).find(
            partial(self.is_active_cell_for_tightend_subvolume, birth, optimal_volume_cells)
        )

    def is_active_cell_for_tightend_subvolume(self, birth, optimal_volume_cells, cell_index):
        return cell_index in optimal_volume_cells and (
            self.index_to_level[birth] + self.epsilon < self.index_to_level[cell_index]
        )


class OptimalVolumeTightenedSubVolumeFinder(object):
    def __init__(self, optimizer_builder, index_to_level, epsilon):
        self.optimizer_builder = optimizer_builder
        self.tsv_finder = TightenedSubVolumeFinder(optimizer_builder.builder_without_retry(), index_to_level, epsilon)

    def find(self, birth, death):
        optimal_volume = self.optimizer_builder.build(birth, death).find()
        if not optimal_volume.success:
            return optimal_volume
        optimal_volume.subvolume = self.tsv_finder.find(birth, death, optimal_volume.cell_indices)
        return optimal_volume

    def to_query_dict(self):
        return dict(self.optimizer_builder.to_query_dict(), **{"query-target": "tightened-subvolume"})


class PointQuery(object):
    def __init__(self, birth_time, death_time, ovfinder, spatial_searcher):
        self.birth_time = birth_time
        self.death_time = death_time
        self.ovfinder = ovfinder
        self.spatial_searcher = spatial_searcher

    def to_dict(self):
        return dict(
            self.ovfinder.to_query_dict(),
            **{
                "birth": self.birth_time,
                "death": self.death_time,
            }
        )

    def execute(self):
        pair_indices, pair = self.spatial_searcher.nearest_pair(self.birth_time, self.death_time)
        result = self.ovfinder.find(*pair_indices)
        result.pair = pair
        return [result]

    @staticmethod
    def valid_args(args):
        return args.x is not None and args.y is not None


class RectangleQuery(object):
    def __init__(self, birth_range, death_range, ovfinder, spatial_searcher, skip_infeasible):
        self.birth_range = birth_range
        self.death_range = death_range
        self.ovfinder = ovfinder
        self.spatial_searcher = spatial_searcher
        self.skip_infeasible = skip_infeasible

    def to_dict(self):
        return dict(
            self.ovfinder.to_query_dict(),
            **{
                "birth-range": self.birth_range,
                "death-range": self.death_range,
                "skip-infeasible": self.skip_infeasible,
            }
        )

    def pairs_in_rectangle(self):
        return self.spatial_searcher.in_rectangle(
            self.birth_range[0], self.birth_range[1], self.death_range[0], self.death_range[1]
        )

    def execute(self):
        results = []
        for pair_indices, pair in self.pairs_in_rectangle():
            result = self.ovfinder.find(*pair_indices)
            result.pair = pair
            if result.success:
                results.append(result)
            elif result.infeasible and self.skip_infeasible:
                results.append(result)
            else:
                raise RuntimeError(result.message)
        return results

    @staticmethod
    def valid_args(args):
        return args.x_range and args.y_range


class SuccessToDictBase(object):
    def __init__(self, index_to_level, death_to_birth):
        self.index_to_level = index_to_level
        self.death_to_birth = death_to_birth

    def children(self, death, success):
        ret = []
        for b, d in success.children(death, self.death_to_birth):
            birth_time = self.index_to_level[b]
            death_time = self.index_to_level[d]
            if birth_time == death_time:
                pass
            ret.append(
                {
                    "birth-time": birth_time,
                    "death-time": death_time,
                    "birth-index": b,
                    "death-index": d,
                }
            )

        return ret

    def basedict(self, success, death):
        birth = self.death_to_birth[death]

        return {
            "birth-time": self.index_to_level[birth],
            "death-time": self.index_to_level[death],
            "birth-index": birth,
            "death-index": death,
            "success": True,
            "tightened-subvolume": self(success.subvolume),
        }


class SuccessToDictAlpha(SuccessToDictBase):
    def __init__(self, index_to_level, coord_resolver, symbol_resolver, death_to_birth):
        super().__init__(index_to_level, death_to_birth)
        self.coord_resolver = coord_resolver
        self.symbol_resolver = symbol_resolver

    @staticmethod
    def from_pdgm(pdgm):
        return SuccessToDictAlpha(
            pdgm.index_to_level, pdgm.alpha_coord_resolver, pdgm.alpha_symbol_resolver, pdgm.death_index_to_birth_index
        )

    def __call__(self, success):
        if success is None:
            return None

        death = max(success.cell_indices)

        return dict(
            self.basedict(success, death),
            **{
                "points": self.coord_resolver.resolve_vertices(success.cell_indices),
                "simplices": self.coord_resolver.resolve_cells(success.cell_indices),
                "boundary": self.coord_resolver.resolve_cells(
                    self.coord_resolver.boundary_cells(success.cell_indices)
                ),
                "boundary-points": self.coord_resolver.resolve_vertices(
                    self.coord_resolver.boundary_cells(success.cell_indices)
                ),
                "points-symbols": self.symbol_resolver.resolve_vertices(success.cell_indices),
                "simplices-symbols": self.symbol_resolver.resolve_cells(success.cell_indices),
                "boundary-symbols": self.symbol_resolver.resolve_cells(
                    self.symbol_resolver.boundary_cells(success.cell_indices)
                ),
                "boundary-points-symbols": self.symbol_resolver.resolve_vertices(
                    self.symbol_resolver.boundary_cells(success.cell_indices)
                ),
                "children": self.children(death, success),
            }
        )


class SuccessToDictAbstract(SuccessToDictBase):
    def __init__(self, index_to_level, abs_resolver, death_to_birth):
        super().__init__(index_to_level, death_to_birth)
        self.abs_resolver = abs_resolver

    def __call__(self, success):
        if success is None:
            return None

        death = max(success.cell_indices)
        return dict(
            self.basedict(success, death),
            **{
                "cells": self.abs_resolver.resolve_cells(success.cell_indices),
                "boundary": self.abs_resolver.resolve_cells(self.abs_resolver.boundary_cells(success.cell_indices)),
                "children": self.children(death, success),
            }
        )

    @staticmethod
    def from_pdgm(pdgm):
        return SuccessToDictAbstract(
            pdgm.index_to_level, pdgm.abstract_geometry_resolver, pdgm.death_index_to_birth_index
        )


class SuccessToDictSimplicial(SuccessToDictBase):
    def __init__(self, index_to_level, symbol_resolver, death_to_birth):
        super().__init__(index_to_level, death_to_birth)
        self.symbol_resolver = symbol_resolver

    def __call__(self, success):
        if success is None:
            return None

        death = max(success.cell_indices)
        return dict(
            self.basedict(success, death),
            **{
                "points-symbols": self.symbol_resolver.resolve_vertices(success.cell_indices),
                "simplices-symbols": self.symbol_resolver.resolve_cells(success.cell_indices),
                "boundary-symbols": self.symbol_resolver.resolve_cells(
                    self.symbol_resolver.boundary_cells(success.cell_indices)
                ),
                "boundary-points-symbols": self.symbol_resolver.resolve_vertices(
                    self.symbol_resolver.boundary_cells(success.cell_indices)
                ),
                "children": self.children(death, success),
            }
        )

    @staticmethod
    def from_pdgm(pdgm):
        return SuccessToDictSimplicial(
            pdgm.index_to_level, pdgm.simplicial_symbol_resolver, pdgm.death_index_to_birth_index
        )


class SuccessToDictCubical(SuccessToDictBase):
    def __init__(self, index_to_level, cube_resolver, death_to_birth):
        super().__init__(index_to_level, death_to_birth)
        self.cube_resolver = cube_resolver

    @staticmethod
    def from_pdgm(pdgm):
        return SuccessToDictCubical(
            pdgm.index_to_level, pdgm.cubical_geometry_resolver, pdgm.death_index_to_birth_index
        )

    def __call__(self, success):
        if success is None:
            return None

        death = max(success.cell_indices)

        return dict(
            self.basedict(success, death),
            **{
                "points": self.cube_resolver.resolve_vertices(success.cell_indices),
                "cubes": self.cube_resolver.resolve_cells(success.cell_indices),
                "boundary": self.cube_resolver.resolve_cells(self.cube_resolver.boundary_cells(success.cell_indices)),
                "boundary-points": self.cube_resolver.resolve_vertices(
                    self.cube_resolver.boundary_cells(success.cell_indices)
                ),
                "children": self.children(death, success),
            }
        )


def failure_to_dict(failure):
    return {
        "birth-time": failure.pair[0],
        "death-time": failure.pair[1],
        "success": False,
        "status": pulp.LpStatus[failure.status],
    }


class OptimalVolumeError(Exception):
    """Base class for expretions in homcloud.optimal_volume"""

    def __init__(self, message, code):
        self.message = message
        self.code = code


class InfeasibleError(OptimalVolumeError):
    """Exception raised for errors for LpstatusInfeasible."""

    def __init__(self, message, code):
        super().__init__(message, code)


def argument_parser():
    p = argparse.ArgumentParser()
    p.description = "Copmutes optimal volumes and variants of stable volumes"
    p.add_argument("-V", "--version", action="version", version=__version__)
    add_argument_for_license(p)

    tp = p.add_argument_group("target")
    tp.add_argument("-d", "--degree", type=int, required=True, help="degree of PH")
    tp.add_argument("-x", type=float, help="birth time of the pair")
    tp.add_argument("-y", type=float, help="death time of the pair")
    tp.add_argument("-X", "--x-range", type=parse_range, help="birth time of the pair")
    tp.add_argument("-Y", "--y-range", type=parse_range, help="death time of the pair")

    op = p.add_argument_group("output options")
    op.add_argument("-j", "--json-output", help="output in json format")
    op.add_argument("-P", "--invoke-paraview", default=False, action="store_true", help="invoke paraview")

    cp = p.add_argument_group("computation parameters")
    cp.add_argument(
        "-T",
        "--type",
        default="optimal-volume",
        help="target type (*optimal-volume*, tightened-volume," " or tightened-subvolume",
    )
    cp.add_argument("-e", "--epsilon", type=float, default=0.0, help="tighened-volume/subvolume epsilon")
    cp.add_argument("-c", "--cutoff-radius", type=float, help="cut-off radius in R^n")
    cp.add_argument("-n", "--retry", type=int, default=1, help="number of retry")
    cp.add_argument("--skip-infeasible", default=False, type=parse_bool, help="skip infeasible (on/*off*)")

    sp = p.add_argument_group("solver parameters")
    sp.add_argument("--solver", help="LP solver name")
    sp.add_argument("-O", "--option", action="append", default=[], help="options for LP solver")

    p.add_argument("input", help="input filename")
    return p


def build_spatial_searcher(diagram):
    return SpatialSearcher(
        list(zip(zip(diagram.birth_indices, diagram.death_indices), zip(diagram.births, diagram.deaths))),
        diagram.births,
        diagram.deaths,
    )


@forwardable()
class Main(object):
    def __init__(self, args):
        self.args = args

    def_delegators("args", "input, degree, cutoff_radius, retry")

    def run(self):
        if self.input_is_pdgm:
            with PDGM.open(self.input, self.degree) as pdgm:
                self.setup_pdgm(pdgm)

        ovfinder = self.build_ovfinder()
        query = self.build_query(ovfinder)
        results = query.execute()

        if self.is_json_output:
            self.write_json(query, results)
        if self.args.invoke_paraview:
            self.invoke_paraview(results)

    @property
    def input_is_pdgm(self):
        return Path(self.input).suffix == ".pdgm"

    @property
    def lp_solver(self):
        return find_lp_solver(self.args.solver, self.args.option)

    @property
    def is_json_output(self):
        return self.args.json_output

    @property
    def points(self):
        return self.coord_resolver.vertices

    def setup_pdgm(self, pdgm):
        if pdgm.filtration_type == "alpha":
            self.setup_alpha_pdgm(pdgm)
        elif pdgm.filtration_type == "abstract":
            self.setup_abstract_pdgm(pdgm)
        elif pdgm.filtration_type == "simplicial":
            self.setup_simplicial_pdgm(pdgm)
        elif pdgm.filtration_type == "cubical":
            self.setup_cubical_pdgm(pdgm)
        else:
            raise RuntimeError("Unknown pdgm type: {}".format(pdgm.filtration_type))
        self.filtration_type = pdgm.filtration_type
        self.spatial_searcher = build_spatial_searcher(pdgm)
        self.index_to_level = pdgm.index_to_level

    def setup_alpha_pdgm(self, pdgm):
        self.optimizer_builder = OptimizerBuilder.from_alpha_pdgm(pdgm, self.cutoff_radius, self.retry, self.lp_solver)
        self.success_to_dict = SuccessToDictAlpha.from_pdgm(pdgm)
        self.coord_resolver = pdgm.alpha_coord_resolver

    def setup_abstract_pdgm(self, pdgm):
        self.optimizer_builder = OptimizerBuilder(self.degree, pdgm.boundary_map_chunk, self.lp_solver)
        self.success_to_dict = SuccessToDictAbstract.from_pdgm(pdgm)

    def setup_simplicial_pdgm(self, pdgm):
        self.optimizer_builder = OptimizerBuilder(self.degree, pdgm.boundary_map_chunk, self.lp_solver)
        self.success_to_dict = SuccessToDictSimplicial.from_pdgm(pdgm)

    def setup_cubical_pdgm(self, pdgm):
        self.optimizer_builder = OptimizerBuilder.from_cubical_pdgm(
            pdgm, self.cutoff_radius, self.retry, self.lp_solver
        )
        self.success_to_dict = SuccessToDictCubical.from_pdgm(pdgm)
        self.cube_resolver = pdgm.cubical_geometry_resolver

    def build_ovfinder(self):
        type = self.args.type
        if type == "optimal-volume":
            return OptimalVolumeFinder(self.optimizer_builder)
        elif type == "tightened-volume":
            return TightenedVolumeFinder(self.optimizer_builder, self.index_to_level, self.args.epsilon)
        elif type == "tightened-subvolume":
            return OptimalVolumeTightenedSubVolumeFinder(
                self.optimizer_builder, self.index_to_level, self.args.epsilon
            )
        else:
            raise RuntimeError("Unknown type: {}".format(type))

    def build_query(self, ovfinder):
        if PointQuery.valid_args(self.args):
            return PointQuery(self.args.x, self.args.y, ovfinder, self.spatial_searcher)
        elif RectangleQuery.valid_args(self.args):
            return RectangleQuery(
                self.args.x_range, self.args.y_range, ovfinder, self.spatial_searcher, self.args.skip_infeasible
            )
        else:
            raise RuntimeError("Invalid query")

    def write_json(self, query, results):
        with open(self.args.json_output, "w") as f:
            json.dump(self.build_jsondict(query, results), f, cls=NumpyEncoder)

    def build_jsondict(self, query, results):
        def result_to_dict(result):
            if result.success:
                return self.success_to_dict(result)
            else:
                return failure_to_dict(result)

        return {"format-version": 2.0, "query": query.to_dict(), "result": [result_to_dict(r) for r in results]}

    def invoke_paraview(self, results):
        if self.filtration_type == "alpha":
            drawer = self.draw_volumes_for_alpha(results)
        elif self.filtration_type == "cubical":
            drawer = self.draw_volumes_for_cubical(results)
        else:
            raise RuntimeError("Paraview visualization is not available")

        drawer.invoke()

    def draw_volumes_for_alpha(self, results):
        drawer = drawer_for_alpha(self.points)
        draw_volumes(drawer, results, self.coord_resolver)
        return drawer

    def draw_volumes_for_cubical(self, results):
        drawer = drawer_for_cubical(self.cube_resolver.shape)
        draw_volumes(drawer, results, self.cube_resolver)
        return drawer


def drawer_for_alpha(points):
    return ParaViewSimplexDrawer(4, points, {"index": None, "isboundary": 0, "issimplex": 0, "issubvolume": 0})


def drawer_for_cubical(shape):
    return ParaViewCubeDrawer(4, shape, {"index": None, "isboundary": 0, "issimplex": 0, "issubvolume": 0})


def draw_volumes(drawer, results, geom_resolver):
    for i, volume in enumerate(results):
        volume.draw(drawer, geom_resolver, drawer.various_colors[0], drawer.various_colors[1], i)


def main(args=None):
    Main(args or argument_parser().parse_args()).run()


if __name__ == "__main__":
    main()
